from ..scanner import Scanner
from ..scanner_utils import is_white_space, is_number

class TokenType:
    Number = 'num'
    Op1 = 'op1'
    Op2 = 'op2'
    Null = 'null'


class Operator:
    Plus = '+'
    Minus = '-'
    Multiply = '*'
    Divide = '/'
    IntDivide = '\\'
    LeftParenthesis = '('
    RightParenthesis = ')'
    Dot = '.'


class ParserState:
    Primary = 1 << 0
    Operator = 1 << 1
    LParen = 1 << 2
    RParen = 1 << 3
    Sign = 1 << 4
    NullaryCall = 1 << 5


class Token:
    __slots__ = ('type', 'value', 'priority')

    def __init__(self, token_type: TokenType, value: int, priority=0):
        self.type = token_type
        self.value = value
        self.priority = priority

class MathExpressionException(Exception):
    def __init__(self, message: str, scanner: Scanner=None):
        if scanner:
            message += ' at column %d of expression' % scanner.pos
        super(MathExpressionException, self).__init__(message)
        self.message = message
        if scanner:
            self.pos = scanner.pos


nullary = Token(TokenType.Null, 0)


def parse(expr: str):
    "Parses given expression in forward direction"
    scanner = Scanner(expr) if isinstance(expr, str) else expr
    priority = 0
    expected = ParserState.Primary | ParserState.LParen | ParserState.Sign
    tokens = []

    while not scanner.eof():
        scanner.eat_while(is_white_space)
        scanner.start = scanner.pos

        if consume_number(scanner):
            if (expected & ParserState.Primary) == 0:
                raise MathExpressionException('Unexpected number', scanner)

            tokens.append(number(scanner.current()))
            expected = ParserState.Operator | ParserState.RParen
        elif is_operator(scanner.peek()):
            ch = scanner.next()
            if is_sign(ch) and (expected & ParserState.Sign):
                if is_negative_sign(ch):
                    tokens.append(op1(ch, priority))
                expected = ParserState.Primary | ParserState.LParen | ParserState.Sign
            else:
                if (expected & ParserState.Operator) == 0:
                    raise MathExpressionException('Unexpected operator', scanner)
                tokens.append(op2(ch, priority))
                expected = ParserState.Primary | ParserState.LParen | ParserState.Sign
        elif scanner.eat(Operator.LeftParenthesis):
            if (expected & ParserState.LParen) == 0:
                raise MathExpressionException('Unexpected "("', scanner)

            priority += 10
            expected = ParserState.Primary | ParserState.LParen | ParserState.Sign | ParserState.NullaryCall
        elif scanner.eat(Operator.RightParenthesis):
            priority -= 10

            if expected & ParserState.NullaryCall:
                tokens.append(nullary)
            elif (expected & ParserState.RParen) == 0:
                raise MathExpressionException('Unexpected ")"', scanner)

            expected = ParserState.Operator | ParserState.RParen | ParserState.LParen
        else:
            raise MathExpressionException('Unknown character', scanner)

    if 0 < priority >= 10:
        raise MathExpressionException('Unmatched "()"', scanner)

    result = order_tokens(tokens)

    if result is None:
        raise MathExpressionException('Parity', scanner)

    return result

def consume_number(scanner: Scanner):
    """
    Consumes number from given stream
    :return `True` if number was consumed
    """
    start = scanner.pos
    if scanner.eat(Operator.Dot) and scanner.eat_while(is_number):
        # short decimal notation: .025
        return True

    if scanner.eat_while(is_number) and (not scanner.eat(Operator.Dot) or scanner.eat_while(is_number)):
        # either integer or decimal: 10, 10.25
        return True

    scanner.pos = start
    return False


def order_tokens(tokens: list):
    """
    Orders parsed tokens (operands and operators) in given array so that they are
    laid off in order of execution
    """
    operators = []
    operands = []
    n_operators = 0

    for t in tokens:
        if t.type == TokenType.Number:
            operands.append(t)
        else:
            n_operators += 1 if t.type == TokenType.Op1 else 2

            while operators:
                if t.priority <= operators[-1].priority:
                    operands.append(operators.pop())
                else:
                    break

            operators.append(t)

    if n_operators + 1 == len(operands) + len(operators):
        operators.reverse()
        return operands + operators

    # Parity
    return None


def number(value: str, priority=0):
    "Number token factory"
    return Token(TokenType.Number, float(value), priority)


def op1(value: str, priority=0):
    "Unary operator factory"
    if value == Operator.Minus:
        priority += 2
    return Token(TokenType.Op1, value, priority)


def op2(value: str, priority=0):
    "Binary operator factory"
    if value == Operator.Multiply:
        priority += 1
    elif value == Operator.Divide or value == Operator.IntDivide:
        priority += 2

    return Token(TokenType.Op2, value, priority)


def is_sign(ch: str):
    return is_positive_sign(ch) or is_negative_sign(ch)


def is_positive_sign(ch: str):
    return ch == Operator.Plus


def is_negative_sign(ch: str):
    return ch == Operator.Minus


def is_operator(ch: str):
    return ch in (Operator.Plus, Operator.Minus, Operator.Multiply, Operator.Divide, Operator.IntDivide)



import re
import collections
from ..css_abbreviation import parse, tokens, CSSValue, FunctionCall


re_property = re.compile(r'^([a-z-]+)(?:\s*:\s*([^\n\r;]+?);*)?$')
opt = {'value': True}


class CSSSnippetType:
    Raw = 'Raw'
    Property = 'Property'


class CSSSnippetRaw:
    __slots__ = ('type', 'key', 'value')

    def __init__(self, key: str, value: str):
        self.type = CSSSnippetType.Raw
        self.key = key
        self.value = value


class CSSSnippetProperty:
    __slots__ = ('type', 'key', 'value', 'property', 'keywords', 'dependencies')

    def __init__(self, key: str, prop: str, value: list, keywords: dict):
        self.type = CSSSnippetType.Property
        self.key = key
        self.property = prop
        self.value = value
        self.keywords = keywords
        self.dependencies = []


def create_snippet(key: str, value: str):
    "Creates structure for holding resolved CSS snippet"
    # A snippet could be a raw text snippet (e.g. arbitrary text string) or a
    # CSS property with possible values separated by `|`.
    # In latter case, we have to parse snippet as CSS abbreviation
    m = re_property.match(value)
    if m:
        keywords = collections.OrderedDict()
        parsed = [parse_value(v) for v in m.group(2).split('|')] if m.group(2) else []

        for item in parsed:
            for css_val in item:
                collect_keywords(css_val, keywords)

        return CSSSnippetProperty(key, m.group(1), parsed, keywords)

    return CSSSnippetRaw(key, value)


def nest(snippets: list):
    """
    Nests more specific CSS properties into shorthand ones, e.g.
    `background-position-x` -> `background-position` -> `background`
    """
    snippets = snippets[:]
    snippets.sort(key=lambda x: x.key)
    stack = []

    # For sorted list of CSS properties, create dependency graph where each
    # shorthand property contains its more specific one, e.g.
    # background -> background-position -> background-position-x
    for cur in filter(is_property, snippets):
        # Check if current property belongs to one from parent stack.
        # Since `snippets` array is sorted, items are perfectly aligned
        # from shorthands to more specific variants
        while stack:
            prev = stack[-1]

            if cur.property.startswith(prev.property) and \
                len(cur.property) > len(prev.property) and \
                cur.property[len(prev.property)] == '-':
                prev.dependencies.append(cur)
                stack.append(cur)
                break

            stack.pop()

        if not stack:
            stack.append(cur)

    return snippets


def parse_value(value: str):
    global opt
    return parse(value.strip(), opt)[0].value


def is_property(snippet):
    return isinstance(snippet, CSSSnippetProperty)


def collect_keywords(css_val: CSSValue, dest: dict):
    for v in css_val.value:
        if isinstance(v, tokens.Literal):
            dest[v.value] = v
        elif isinstance(v, FunctionCall):
            dest[v.name] = v
        elif isinstance(v, tokens.Field):
            # Create literal from field, if available
            value = v.name.strip()
            if value:
                dest[value] = tokens.Literal(value)



import unittest
import sys

sys.path.append('../')
from emmet.output_stream import OutputStream, tag_name, attr_name, self_close, is_inline
from emmet.config import Config

class TestOutputStream(unittest.TestCase):
    def test_stream(self):
        conf = Config({'options': {'output.baseIndent': '>>'}})
        out = OutputStream(conf.options)

        out.push('aaa')
        self.assertEqual(out.value, 'aaa')
        self.assertEqual(out.line, 0)
        self.assertEqual(out.column, 3)
        self.assertEqual(out.offset, 3)

        out.push_string('bbb')
        self.assertEqual(out.value, 'aaabbb')
        self.assertEqual(out.line, 0)
        self.assertEqual(out.column, 6)
        self.assertEqual(out.offset, 6)

        # Add text with newlines
        out.push_string('ccc\nddd')
        self.assertEqual(out.value, 'aaabbbccc\n>>ddd')
        self.assertEqual(out.line, 1)
        self.assertEqual(out.column, 5)
        self.assertEqual(out.offset, 15)

        # Add newline with indent
        out.level += 1
        out.push_newline(True)
        self.assertEqual(out.value, 'aaabbbccc\n>>ddd\n>>\t')
        self.assertEqual(out.line, 2)
        self.assertEqual(out.column, 3)
        self.assertEqual(out.offset, 19)

    def test_profile_tag_name(self):
        as_is = Config({ 'options': { 'output.tagCase': '' } })
        upper = Config({ 'options': { 'output.tagCase': 'upper' } })
        lower = Config({ 'options': { 'output.tagCase': 'lower' } })

        self.assertEqual(tag_name('Foo', as_is), 'Foo')
        self.assertEqual(tag_name('bAr', as_is), 'bAr')

        self.assertEqual(tag_name('Foo', upper), 'FOO')
        self.assertEqual(tag_name('bAr', upper), 'BAR')

        self.assertEqual(tag_name('Foo', lower), 'foo')
        self.assertEqual(tag_name('bAr', lower), 'bar')

    def test_attribute_name(self):
        as_is = Config({ 'options': { 'output.attributeCase': '' } })
        upper = Config({ 'options': { 'output.attributeCase': 'upper' } })
        lower = Config({ 'options': { 'output.attributeCase': 'lower' } })

        self.assertEqual(attr_name('Foo', as_is), 'Foo')
        self.assertEqual(attr_name('bAr', as_is), 'bAr')

        self.assertEqual(attr_name('Foo', upper), 'FOO')
        self.assertEqual(attr_name('bAr', upper), 'BAR')

        self.assertEqual(attr_name('Foo', lower), 'foo')
        self.assertEqual(attr_name('bAr', lower), 'bar')

    def test_self_close(self):
        html = Config({ 'options': { 'output.selfClosingStyle': 'html' } })
        xhtml = Config({ 'options': { 'output.selfClosingStyle': 'xhtml' } })
        xml = Config({ 'options': { 'output.selfClosingStyle': 'xml' } })

        self.assertEqual(self_close(html), '')
        self.assertEqual(self_close(xhtml), ' /')
        self.assertEqual(self_close(xml), '/')

    def test_inline_elements(self):
        config = Config()
        self.assertEqual(is_inline('a', config), True)
        self.assertEqual(is_inline('b', config), True)
        self.assertEqual(is_inline('c', config), False)



from ..html_matcher import scan, attributes, ScannerOptions, ElementType, AttributeToken
from .utils import push_range, SelectItemModel, token_list

class ContextTag:
    __slots__ = ('name', 'type', 'start', 'end', 'attributes')

    def __init__(self, name: str, elem_type: ElementType, start: int, end: int, attrs: list=None):
        self.name = name
        self.type = elem_type
        self.start = start
        self.end = end
        self.attributes = attrs

    def to_json(self):
        "Returns JSON representation of current object"
        result = {
            'name': self.name,
            'type': self.type,
            'start': self.start,
            'end': self.end
        }

        if self.attributes:
            result['attributes'] = [attr.to_json() for attr in self.attributes]

        return result


def get_open_tag(code: str, pos: int) -> ContextTag:
    """
    Check if there’s open or self-closing tag under given `pos` location in source code.
    If found, returns its name, range in source and parsed attributes
    """
    opt = ScannerOptions()
    tag = []
    tag.append(None)

    # Find open or self-closing tag, closest to given position
    def scan_callback(name: str, elem_type: ElementType, start: int, end: int):
        if start < pos < end:
            tag[0] = ContextTag(name, elem_type, start, end)
            if elem_type in (ElementType.Open, ElementType.SelfClose):
                tag[0].attributes = shift_attribute_ranges(attributes(code[start:end], name), start)

            return False
        if end > pos:
            return False

    scan(code, scan_callback, opt.special)

    return tag[0]


def select_item_html(code: str, pos: int, is_prev=False, options: dict=None) -> SelectItemModel:
    "Returns list of ranges for Select Next/Previous Item action"
    if is_prev:
        return select_previous_item(code, pos, options)
    return select_next_item(code, pos, options)


def select_next_item(code: str, pos: int, options: dict=None) -> SelectItemModel:
    "Returns list of ranges for Select Next Item action"
    opt = ScannerOptions(options)
    result = []
    result.append(None)

    # Find open or self-closing tag, closest to given position
    def scan_callback(name, elem_type, start, end):
        if elem_type in (ElementType.Open, ElementType.SelfClose) and end > pos:
            # Found open or self-closing tag
            result[0] = get_tag_selection_model(code, name, start, end)
            return False
    scan(code, scan_callback, opt.special)
    return result[0]


def select_previous_item(code: str, pos: int, options: dict=None) -> SelectItemModel:
    "Returns list of ranges for Select Previous Item action"
    opt = ScannerOptions(options)
    last = {
        'name': '',
        'type': None,
        'start': -1,
        'end': -1
    }

    # We should find the closest open or self-closing tag left to given `pos`.
    def scan_callback(name: str, elem_type: ElementType, start: int, end: int):
        if start >= pos:
            return False

        if elem_type in (ElementType.Open, ElementType.SelfClose):
            # Found open or self-closing tag
            last['name'] = name
            last['type'] = elem_type
            last['start'] = start
            last['end'] = end

    scan(code, scan_callback, opt.special)

    if last['type'] is not None:
        return get_tag_selection_model(code, last['name'], last['start'], last['end'])


def get_tag_selection_model(code: str, name: str, start: int, end: int) -> SelectItemModel:
    """
    Parses open or self-closing tag in `start:end` range of `code` and returns its
    model for selecting items
    :param code Document source code
    :param name Name of matched tag
    """
    # Add tag name range
    ranges = [(start + 1, start + 1 + len(name))]

    # Parse and add attributes ranges
    tag_src = code[start:end]
    for attr in attributes(tag_src, name):
        if attr.value is not None:
            # Attribute with value
            push_range(ranges, (start + attr.name_start, start + attr.value_end))

            # Add (unquoted) value range
            val = value_range(attr)
            if val[0] != val[1]:
                push_range(ranges, (start + val[0], start + val[1]))

                if attr.name == 'class':
                    # For class names, split value into space-separated tokens
                    for token in token_list(tag_src[val[0]:val[1]], start + val[0]):
                        push_range(ranges, token)
        else:
            # Attribute without value (boolean)
            push_range(ranges, (start + attr.name_start, start + attr.name_end))

    return SelectItemModel(start, end, ranges)


def value_range(attr: AttributeToken) -> tuple:
    "Returns value range of given attribute. Value range is unquoted."
    value = attr.value
    ch = value[0]
    last_ch = value[-1]
    if ch == '"' or ch == '\'':
        return (
            attr.value_start + 1,
            attr.value_end - (1 if last_ch == ch else 0)
        )

    if ch == '{' and last_ch == '}':
        return (
            attr.value_start + 1,
            attr.value_end - 1
        )

    return (attr.value_start, attr.value_end)

def shift_attribute_ranges(attrs: list, offset: int):
    for attr in attrs:
        attr.name_start += offset
        attr.name_end += offset
        if attr.value is not None:
            attr.value_start += offset
            attr.value_end += offset
    return attrs



snippets = {
    "tm|tmatch": "xsl:template[match mode]",
    "tn|tname": "xsl:template[name]",
    "call": "xsl:call-template[name]",
    "ap": "xsl:apply-templates[select mode]",
    "api": "xsl:apply-imports",
    "imp": "xsl:import[href]",
    "inc": "xsl:include[href]",
    "ch": "xsl:choose",
    "wh|xsl:when": "xsl:when[test]",
    "ot": "xsl:otherwise",
    "if": "xsl:if[test]",
    "par": "xsl:param[name]",
    "pare": "xsl:param[name select]",
    "var": "xsl:variable[name]",
    "vare": "xsl:variable[name select]",
    "wp": "xsl:with-param[name select]",
    "key": "xsl:key[name match use]",
    "elem": "xsl:element[name]",
    "attr": "xsl:attribute[name]",
    "attrs": "xsl:attribute-set[name]",
    "cp": "xsl:copy[select]",
    "co": "xsl:copy-of[select]",
    "val": "xsl:value-of[select]",
    "for|each": "xsl:for-each[select]",
    "tex": "xsl:text",
    "com": "xsl:comment",
    "msg": "xsl:message[terminate=no]",
    "fall": "xsl:fallback",
    "num": "xsl:number[value]",
    "nam": "namespace-alias[stylesheet-prefix result-prefix]",
    "pres": "xsl:preserve-space[elements]",
    "strip": "xsl:strip-space[elements]",
    "proc": "xsl:processing-instruction[name]",
    "sort": "xsl:sort[select order]",
    "choose": "xsl:choose>xsl:when+xsl:otherwise",
    "xsl": "!!!+xsl:stylesheet[version=1.0 xmlns:xsl=http://www.w3.org/1999/XSL/Transform]>{\n|}",
    "!!!": "{<?xml version=\"1.0\" encoding=\"UTF-8\"?>}"
}



import unittest
import sys

sys.path.append('../')

from emmet.extract_abbreviation import extract_abbreviation, is_at_html_tag, ExtractedAbbreviation
from emmet.extract_abbreviation.reader import BackwardScanner
from emmet.extract_abbreviation.is_html import consume_quoted

def extract(abbr: str, options={}):
    caret_pos = abbr.find('|')
    if caret_pos != -1:
        abbr = abbr[0:caret_pos] + abbr[caret_pos + 1:]
    else:
        caret_pos = None

    return extract_abbreviation(abbr, caret_pos, options)


def result(abbreviation: str, location: int, start=None):
    if start is None: start = location
    return ExtractedAbbreviation(abbreviation, location, start, location + len(abbreviation))


class TestExtractAbbreviation(unittest.TestCase):
    def test_basic(self):
        self.assertEqual(extract('.bar'), result('.bar', 0))
        self.assertEqual(extract('.foo .bar'), result('.bar', 5))
        self.assertEqual(extract('.foo @bar'), result('@bar', 5))
        self.assertEqual(extract('.foo img/'), result('img/', 5))
        self.assertEqual(extract('текстdiv'), result('div', 5))
        self.assertEqual(extract('foo div[foo="текст" bar=текст2]'), result('div[foo="текст" bar=текст2]', 4))

        # https://github.com/emmetio/emmet/issues/577
        self.assertEqual(
            extract('table>(tr.prefix-intro>td*1)+(tr.prefix-pro-con>th*1+td*3)+(tr.prefix-key-specs>th[colspan=2]*1+td[colspan=2]*3)+(tr.prefix-key-find-online>th[colspan=2]*1+td*2)'),
            result('table>(tr.prefix-intro>td*1)+(tr.prefix-pro-con>th*1+td*3)+(tr.prefix-key-specs>th[colspan=2]*1+td[colspan=2]*3)+(tr.prefix-key-find-online>th[colspan=2]*1+td*2)', 0))

    def test_abbreviation_with_operators(self):
        self.assertEqual(extract('a foo+bar.baz'), result('foo+bar.baz', 2))
        self.assertEqual(extract('a foo>bar+baz*3'), result('foo>bar+baz*3', 2))

    def test_abbreviation_with_attributes(self):
        self.assertEqual(extract('a foo[bar|]'), result('foo[bar]', 2))
        self.assertEqual(extract('a foo[bar="baz" a b]'), result('foo[bar="baz" a b]', 2))
        self.assertEqual(extract('foo bar[a|] baz'), result('bar[a]', 4))

    def test_tag(self):
        self.assertEqual(extract('<foo>bar[a b="c"]>baz'), result('bar[a b="c"]>baz', 5))
        self.assertEqual(extract('foo>bar'), result('foo>bar', 0))
        self.assertEqual(extract('<foo>bar'), result('bar', 5))
        self.assertEqual(extract('<foo>bar[a="d" b="c"]>baz'), result('bar[a="d" b="c"]>baz', 5))

    def test_stylesheet_abbreviation(self):
        self.assertEqual(extract('foo{bar|}'), result('foo{bar}', 0))
        self.assertEqual(extract('foo{bar|}', { 'type': 'stylesheet' }), result('bar', 4))

    def test_prefixed_extract(self):
        self.assertEqual(extract('<foo>bar[a b="c"]>baz'), result('bar[a b="c"]>baz', 5))
        self.assertEqual(extract('<foo>bar[a b="c"]>baz', { 'prefix': '<' }), result('foo>bar[a b="c"]>baz', 1, 0))
        self.assertEqual(extract('<foo>bar[a b="<"]>baz', { 'prefix': '<' }), result('foo>bar[a b="<"]>baz', 1, 0))
        self.assertEqual(extract('<foo>bar{<}>baz', { 'prefix': '<' }), result('foo>bar{<}>baz', 1, 0))

        # Multiple prefix characters
        self.assertEqual(extract('foo>>>bar[a b="c"]>baz', { 'prefix': '>>>' }), result('bar[a b="c"]>baz', 6, 3))

        # Absent prefix
        self.assertEqual(extract('<foo>bar[a b="c"]>baz', {'prefix': '&&'}), None)

    def test_brackets_inside_curly_braces(self):
        self.assertEqual(extract('foo div{[}+a{}'), result('div{[}+a{}', 4))
        self.assertEqual(extract('div{}}'), None)
        self.assertEqual(extract('div{{}'), result('{}', 4))

    def test_html(self):
        html = lambda s: is_at_html_tag(BackwardScanner(s))

        # simple tag
        self.assertTrue(html('<div>'))
        self.assertTrue(html('<div/>'))
        self.assertTrue(html('<div />'))
        self.assertTrue(html('</div>'))

        # tag with attributes
        self.assertTrue(html('<div foo="bar">'))
        self.assertTrue(html('<div foo=bar>'))
        self.assertTrue(html('<div foo>'))
        self.assertTrue(html('<div a="b" c=d>'))
        self.assertTrue(html('<div a=b c=d>'))
        self.assertTrue(html('<div a=^b$ c=d>'))
        self.assertTrue(html('<div a=b c=^%d]$>'))
        self.assertTrue(html('<div title=привет>'))
        self.assertTrue(html('<div title=привет123>'))
        self.assertTrue(html('<foo-bar>'))

        # invalid tags
        self.assertFalse(html('div>'))
        self.assertFalse(html('<div'))
        self.assertFalse(html('<div привет>'))
        self.assertFalse(html('<div =bar>'))
        self.assertFalse(html('<div foo=>'))
        self.assertFalse(html('[a=b c=d]>'))
        self.assertFalse(html('div[a=b c=d]>'))

    def test_consume_quotes(self):
        s = BackwardScanner(' "foo"')
        self.assertTrue(consume_quoted(s))
        self.assertEqual(s.pos, 1)

        s = BackwardScanner('"foo"')
        self.assertTrue(consume_quoted(s))
        self.assertEqual(s.pos, 0)

        s = BackwardScanner('""')
        self.assertTrue(consume_quoted(s))
        self.assertEqual(s.pos, 0)

        s = BackwardScanner('"a\\\"b"')
        self.assertTrue(consume_quoted(s))
        self.assertEqual(s.pos, 0)

        # don’t eat anything
        s = BackwardScanner('foo')
        self.assertFalse(consume_quoted(s))
        self.assertEqual(s.pos, 3)



import unittest
import sys

sys.path.append('../')

from emmet import expand

def field(index: int, placeholder: str, **kwargs):
    if placeholder:
        return '${%d:%s}' % (index, placeholder)
    return '${%d}' % index


class TestExpandMarkup(unittest.TestCase):
    def test_basic(self):
        self.assertEqual(expand('input[value="text$"]*2'), '<input type="text" value="text1"><input type="text" value="text2">')
        self.assertEqual(expand('ul>.item$*2'), '<ul>\n\t<li class="item1"></li>\n\t<li class="item2"></li>\n</ul>')

        # insert text into abbreviation
        self.assertEqual(expand('ul>.item$*', { 'text': ['foo', 'bar'] }), '<ul>\n\t<li class="item1">foo</li>\n\t<li class="item2">bar</li>\n</ul>')

        # Wrap with Abbreviation, skip empty lines
        self.assertEqual(
            expand('ul>.item$*', {'text': ['foo', '', 'bar', '', '']}),
            '<ul>\n\t<li class="item1">foo</li>\n\t<li class="item2">bar</li>\n</ul>'
        )

        # insert TextMate-style fields/tabstops in output
        self.assertEqual(expand('ul>.item$*2', {
            'options': {
                'output.field': field
            }
        }), '<ul>\n\t<li class="item1">${1}</li>\n\t<li class="item2">${2}</li>\n</ul>')

        # https://github.com/emmetio/emmet/issues/725
        self.assertEqual(expand('textarea'), '<textarea name="" id=""></textarea>')

    def test_abbreviations(self):
        snippets = {
            'test': 'test[!foo bar. baz={}]'
        }
        opt = { 'snippets': snippets }
        reverse = {
            'options': { 'output.reverseAttributes': True },
            'snippets': snippets
        }

        self.assertEqual(expand('a.test'), '<a href="" class="test"></a>')
        self.assertEqual(expand('a.test', reverse), '<a class="test" href=""></a>')

        self.assertEqual(expand('test', opt), '<test bar="bar" baz={}></test>')
        self.assertEqual(expand('test[foo]', opt), '<test bar="bar" baz={}></test>')
        self.assertEqual(expand('test[baz=a foo=1]', opt), '<test foo="1" bar="bar" baz={a}></test>')

        # Apply attributes in reverse order
        self.assertEqual(expand('test', reverse), '<test bar="bar" baz={}></test>')
        self.assertEqual(expand('test[foo]', reverse), '<test bar="bar" baz={}></test>')
        self.assertEqual(expand('test[baz=a foo=1]', reverse), '<test baz={a} foo="1" bar="bar"></test>')

    def test_expressions(self):
        self.assertEqual(expand('span{{foo}}'), '<span>{foo}</span>')
        self.assertEqual(expand('span{foo}'), '<span>foo</span>')
        self.assertEqual(expand('span[foo={bar}]'), '<span foo={bar}></span>')
        self.assertEqual(expand('span[foo={{bar}}]'), '<span foo={{bar}}></span>')

    def test_numbering(self):
        self.assertEqual(expand('ul>li.item$@-*5'), '<ul>\n\t<li class="item5"></li>\n\t<li class="item4"></li>\n\t<li class="item3"></li>\n\t<li class="item2"></li>\n\t<li class="item1"></li>\n</ul>')

    def test_syntax(self):
        self.assertEqual(expand('ul>.item$*2', { 'syntax': 'html' }), '<ul>\n\t<li class="item1"></li>\n\t<li class="item2"></li>\n</ul>')
        self.assertEqual(expand('ul>.item$*2', { 'syntax': 'slim' }), 'ul\n\tli.item1 \n\tli.item2 ')
        self.assertEqual(expand('xsl:variable[name=a select=b]>div', { 'syntax': 'xsl' }), '<xsl:variable name="a">\n\t<div></div>\n</xsl:variable>')

    def test_custom_profile(self):
        self.assertEqual(expand('img'), '<img src="" alt="">')
        self.assertEqual(expand('img', { 'options': { 'output.selfClosingStyle': 'xhtml' } }), '<img src="" alt="" />')

    def test_custom_variables(self):
        variables = { 'charset': 'ru-RU' }
        self.assertEqual(expand('[charset=${charset}]{${charset}}'), '<div charset="UTF-8">UTF-8</div>')
        self.assertEqual(expand('[charset=${charset}]{${charset}}', { 'variables': variables }), '<div charset="ru-RU">ru-RU</div>')

    def test_custom_snippets(self):
        snippets = {
            'link': 'link[foo=bar href]/',
            'foo': '.foo[bar=baz]',
            'repeat': 'div>ul>li{Hello World}*3'
        }

        self.assertEqual(expand('foo', { 'snippets': snippets }), '<div class="foo" bar="baz"></div>')

        # `link:css` depends on `link` snippet so changing it will result in
        # altered `link:css` result
        self.assertEqual(expand('link:css'), '<link rel="stylesheet" href="style.css">')
        self.assertEqual(expand('link:css', { 'snippets': snippets }), '<link foo="bar" href="style.css">')

        # https://github.com/emmetio/emmet/issues/468
        self.assertEqual(expand('repeat', { 'snippets': snippets }), '<div>\n\t<ul>\n\t\t<li>Hello World</li>\n\t\t<li>Hello World</li>\n\t\t<li>Hello World</li>\n\t</ul>\n</div>')

    def formatter_options(self):
        self.assertEqual(expand('ul>.item$*2'), '<ul>\n\t<li class="item1"></li>\n\t<li class="item2"></li>\n</ul>')
        self.assertEqual(expand('ul>.item$*2', { 'options': { 'comment.enabled': True } }),
            '<ul>\n\t<li class="item1"></li>\n\t<!-- /.item1 -->\n\t<li class="item2"></li>\n\t<!-- /.item2 -->\n</ul>')

        self.assertEqual(expand('div>p'), '<div>\n\t<p></p>\n</div>')
        self.assertEqual(expand('div>p', { 'options': { 'output.formatLeafNode': True } }), '<div>\n\t<p>\n\t\t\n\t</p>\n</div>')

    def test_jsx(self):
        config = { 'syntax': 'jsx' }
        self.assertEqual(expand('div#foo.bar', config), '<div id="foo" className="bar"></div>')
        self.assertEqual(expand('label[for=a]', config), '<label htmlFor="a"></label>')
        self.assertEqual(expand('Foo.Bar', config), '<Foo.Bar></Foo.Bar>')
        self.assertEqual(expand('div.{theme.style}', config), '<div className={theme.style}></div>')

    def test_override_attributes(self):
        config = { 'syntax': 'jsx' }
        self.assertEqual(expand('.bar', config), '<div className="bar"></div>');
        self.assertEqual(expand('..bar', config), '<div styleName={styles.bar}></div>')
        self.assertEqual(expand('..foo-bar', config), '<div styleName={styles[\'foo-bar\']}></div>')

        self.assertEqual(expand('.foo', { 'syntax': 'vue' }), '<div class="foo"></div>')
        self.assertEqual(expand('..foo', { 'syntax': 'vue' }), '<div :class="foo"></div>')

    def test_wrap_with_abbreviation(self):
        self.assertEqual(
            expand('img[src="$#"]*', {'text': ['foo.jpg', 'bar.jpg']}),
            '<img src="foo.jpg" alt=""><img src="bar.jpg" alt="">')
        self.assertEqual(expand('div>ul', { 'text': ['<div>line1</div>\n<div>line2</div>'] }),
            '<div>\n\t<ul>\n\t\t<div>line1</div>\n\t\t<div>line2</div>\n\t</ul>\n</div>')

        self.assertEqual(expand('a', {'text': 'foo'}), '<a href="">foo</a>')
        self.assertEqual(expand('a', {'text': 'emmet//io'}), '<a href="">emmet//io</a>')
        self.assertEqual(expand('a', {'text': 'http://emmet.io'}), '<a href="http://emmet.io">http://emmet.io</a>')
        self.assertEqual(expand('a', {'text': '//emmet.io'}), '<a href="//emmet.io">//emmet.io</a>')
        self.assertEqual(expand('a', {'text': 'www.emmet.io'}), '<a href="http://www.emmet.io">www.emmet.io</a>')
        self.assertEqual(expand('a', {'text': 'emmet.io'}), '<a href="">emmet.io</a>')
        self.assertEqual(expand('a', {'text': 'info@emmet.io'}), '<a href="mailto:info@emmet.io">info@emmet.io</a>')
        self.assertEqual(expand('a', {'text': 'uSeR@myLongDomainName.com'}), '<a href="mailto:uSeR@myLongDomainName.com">uSeR@myLongDomainName.com</a>')

        self.assertEqual(expand('p', {'text': 'foo\nbar'}), '<p>\n\tfoo\n\tbar\n</p>')
        self.assertEqual(expand('p', {'text': '<div>foo</div>'}), '<p>\n\t<div>foo</div>\n</p>')

        self.assertEqual(expand('a', { 'text': 'https://www.google.it' }), '<a href="https://www.google.it">https://www.google.it</a>')
        self.assertEqual(expand('a', { 'text': 'www.google.it' }), '<a href="http://www.google.it">www.google.it</a>')
        self.assertEqual(expand('a', { 'text': 'google.it' }), '<a href="">google.it</a>')
        self.assertEqual(expand('a', { 'text': 'test here' }), '<a href="">test here</a>')
        self.assertEqual(expand('a', { 'text': 'test@domain.com' }), '<a href="mailto:test@domain.com">test@domain.com</a>')
        self.assertEqual(expand('a', { 'text': 'test here test@domain.com' }), '<a href="">test here test@domain.com</a>')
        self.assertEqual(expand('a', { 'text': 'test here www.domain.com' }), '<a href="">test here www.domain.com</a>')

        self.assertEqual(expand('a[href=]', { 'text': 'https://www.google.it' }), '<a href="https://www.google.it">https://www.google.it</a>')
        self.assertEqual(expand('a[href=]', { 'text': 'www.google.it' }), '<a href="http://www.google.it">www.google.it</a>')
        self.assertEqual(expand('a[href=]', { 'text': 'google.it' }), '<a href="">google.it</a>')
        self.assertEqual(expand('a[href=]', { 'text': 'test here' }), '<a href="">test here</a>')
        self.assertEqual(expand('a[href=]', { 'text': 'test@domain.com' }), '<a href="mailto:test@domain.com">test@domain.com</a>')
        self.assertEqual(expand('a[href=]', { 'text': 'test here test@domain.com' }), '<a href="">test here test@domain.com</a>')
        self.assertEqual(expand('a[href=]', { 'text': 'test here www.domain.com' }), '<a href="">test here www.domain.com</a>')
        self.assertEqual(expand('a[class=here]', { 'text': 'test@domain.com' }), '<a href="mailto:test@domain.com" class="here">test@domain.com</a>')
        self.assertEqual(expand('a.here', { 'text': 'www.domain.com' }), '<a href="http://www.domain.com" class="here">www.domain.com</a>')
        self.assertEqual(expand('a[class=here]', { 'text': 'test here test@domain.com' }), '<a href="" class="here">test here test@domain.com</a>')
        self.assertEqual(expand('a.here', { 'text': 'test here www.domain.com' }), '<a href="" class="here">test here www.domain.com</a>')

        self.assertEqual(expand('a[href="www.google.it"]', { 'text': 'test' }), '<a href="www.google.it">test</a>')
        self.assertEqual(expand('a[href="www.example.com"]', { 'text': 'www.google.it' }), '<a href="www.example.com">www.google.it</a>')

    def test_class_names(self):
        self.assertEqual(expand('div.foo/'), '<div class="foo">')
        self.assertEqual(expand('div.foo1/2'), '<div class="foo1/2"></div>')
        self.assertEqual(expand('div.foo.1/2'), '<div class="foo 1/2"></div>')


class TestExpandPug(unittest.TestCase):
    def test_basics(self):
        self.assertEqual(expand('!', { 'syntax': 'pug' }),
            'doctype html\nhtml(lang="en")\n\thead\n\t\tmeta(charset="UTF-8")\n\t\tmeta(name="viewport", content="width=device-width, initial-scale=1.0")\n\t\ttitle Document\n\tbody ')



import re
from .walk import walk, WalkState
from .utils import caret, is_inline_element, is_snippet, push_tokens, should_output_attribute
from .comment import comment_node_before, comment_node_after, CommentWalkState
from ...abbreviation import Abbreviation, AbbreviationNode, AbbreviationAttribute
from ...abbreviation.tokenizer.tokens import Field
from ...config import Config
from ...output_stream import tag_name, self_close, attr_name, is_boolean_attribute, attr_quote, is_inline, expression_start, expression_end
from ...list_utils import some, find_index, get_item

re_html_tag = re.compile(r'<([\w\-:]+)[\s>]')
re_prop_key = re.compile(r'^[a-zA-Z_$][\w_$]*$')
reserved_keywords = set([
    'for', 'while', 'of', 'async', 'await', 'const', 'let', 'var', 'continue',
    'break', 'debugger', 'do', 'export', 'import', 'in', 'instanceof', 'new', 'return',
    'switch', 'this', 'throw', 'try', 'catch', 'typeof', 'void', 'with', 'yield'
])

class HTMLWalkState(WalkState):
    __slots__ = ('comment')


def html(abbr: Abbreviation, config: Config):
    state = HTMLWalkState(config)
    state.comment = CommentWalkState(config)
    walk(abbr, element, state)
    return state.out.value


def element(node: AbbreviationNode, index: int, items: list, state: HTMLWalkState, walk_next: callable):
    out = state.out
    config = state.config
    fmt = should_format(node, index, items, state)

    # Pick offset level for current node
    level = get_indent(state)
    out.level += level

    if fmt: out.push_newline(True)

    if node.name:
        name = tag_name(node.name, config)
        comment_node_before(node, state)
        out.push_string('<%s' % name)

        if node.attributes:
            for attr in node.attributes:
                if should_output_attribute(attr):
                    push_attribute(attr, state)

        if node.self_closing and not node.children and not node.value:
            out.push_string('%s>' % self_close(config))
        else:
            out.push_string('>')

            if not push_snippet(node, state, walk_next):
                if node.value:
                    inner_format = some(has_newline, node.value) or starts_with_block_tag(node.value, config)
                    if inner_format:
                        out.level += 1
                        out.push_newline(out.level)
                    push_tokens(node.value, state)
                    if inner_format:
                        out.level -= 1
                        out.push_newline(out.level)

                _next(node.children, walk_next)

                if not node.value and not node.children:
                    inner_format = config.options.get('output.formatLeafNode') or \
                        node.name in config.options.get('output.formatForce', [])
                    if inner_format:
                        out.level += 1
                        out.push_newline(out.level)
                    push_tokens(caret, state)
                    if inner_format:
                        out.level -= 1
                        out.push_newline(out.level)

            out.push_string('</%s>' % name)
            comment_node_after(node, state)
    elif not push_snippet(node, state, walk_next) and node.value:
        # A text-only node (snippet)
        push_tokens(node.value, state)
        _next(node.children, walk_next)

    if fmt and index == len(items) - 1 and state.parent:
        offset = 0 if is_snippet(state.parent) else 1
        out.push_newline(out.level - offset)

    out.level -= level


def push_attribute(attr: AbbreviationAttribute, state: WalkState):
    "Outputs given attribute’s content into output stream"
    out = state.out
    config = state.config

    if attr.name:
        attributes = config.options.get('markup.attributes')
        value_prefix = config.options.get('markup.valuePrefix')

        name = attr.name
        value = attr.value
        l_quote = attr_quote(attr, config, True)
        r_quote = attr_quote(attr, config, False)

        if attributes:
            name = get_multi_value(name, attributes, attr.multiple) or name

        name = attr_name(name, config)
        prefix = get_multi_value(attr.name, value_prefix, attr.multiple) if value_prefix else None

        if prefix and value and len(value) == 1 and isinstance(value[0], str):
            # Add given prefix in object notation
            val = value[0]
            fmt = '%s.%s' if is_prop_key(val) else "%s['%s']"
            value = [fmt % (prefix, val)]

            if config.options.get('jsx.enabled'):
                l_quote = expression_start
                r_quote = expression_end

        if is_boolean_attribute(attr, config) and not value:
            # If attribute value is omitted and it’s a boolean value, check for
            # `compactBoolean` option: if it’s disabled, set value to attribute name
            # (XML style)
            if not config.options.get('output.compactBoolean'):
                value = [name]
        elif not value:
            value = caret

        out.push_string(' %s' % name)
        if value:
            out.push_string('=%s' % l_quote)
            push_tokens(value, state)
            out.push_string(r_quote)
        elif config.options.get('output.selfClosingStyle') != 'html':
            out.push_string('=%s%s' % (l_quote, r_quote))


def push_snippet(node: AbbreviationNode, state: WalkState, walk_next: callable):
    if node.value and node.children:
        # We have a value and child nodes. In case if value contains fields,
        # we should output children as a content of first field
        field_ix = find_index(is_field, node.value)
        if field_ix != -1:
            push_tokens(node.value[0:field_ix], state)
            line = state.out.line
            pos = field_ix + 1

            _next(node.children, walk_next)

            # If there was a line change, trim leading whitespace for better result
            if state.out.line != line and isinstance(get_item(node.value, pos), str):
                state.out.push_string(get_item(node.value, pos).lstrip())
                pos += 1

            push_tokens(node.value[pos:], state)
            return True

    return False


def should_format(node: AbbreviationNode, index: int, items: list, state: WalkState):
    "Check if given node should be formatted in its parent context"
    parent = state.parent
    config = state.config

    if not config.options.get('output.format'):
        return False

    if index == 0 and not parent:
        # Do not format very first node
        return False

    # Do not format single child of snippet
    if parent and is_snippet(parent) and len(items) == 1:
        return False

    if is_snippet(node):
        # Adjacent text-only/snippet nodes
        fmt = is_snippet(get_item(items, index - 1)) or is_snippet(get_item(items, index + 1)) or \
            some(has_newline, node.value) or \
            (some(is_field, node.value) and node.children)

        if fmt:
            return True

    if is_inline(node, config):
        # Check if inline node is the next sibling of block-level node
        if index == 0:
            # First node in parent: format if it’s followed by a block-level element
            for item in items:
                if not is_inline(item, config):
                    return True
        elif not is_inline(items[index - 1], config):
            # Node is right after block-level element
            return True

        if config.options.get('output.inlineBreak'):
            # check for adjacent inline elements before and after current element
            adjacent_inline = 1
            before = index - 1
            after = index + 1

            while before >= 0 and is_inline_element(items[before], config):
                adjacent_inline += 1
                before -= 1

            while after < len(items) and is_inline_element(items[after], config):
                adjacent_inline += 1
                after += 1

            if adjacent_inline >= config.options.get('output.inlineBreak'):
                return True

        # Edge case: inline node contains node that should receive formatting
        for i, child in enumerate(node.children):
            if should_format(child, i, node.children, state):
                return True

        return False

    return True


def get_indent(state: WalkState):
    "Returns indentation offset for given node"
    parent = state.parent

    if not parent or is_snippet(parent) or (parent.name and parent.name in state.config.options.get('output.formatSkip')):
        return 0

    return 1


def has_newline(value):
    "Check if given node value contains newlines"
    return '\r' in value or '\n' in value if isinstance(value, str) else False


def starts_with_block_tag(value: list, config: Config) -> bool:
    "Check if given node value starts with block-level tag"
    if value and isinstance(value[0], str):
        m = re_html_tag.match(value[0])

        if m and m.group(1).lower() not in config.options.get('inlineElements'):
            return True

    return False

def _next(items: list, walk_next: callable):
    for i, item in enumerate(items):
        walk_next(item, i, items)


def is_field(item):
    return isinstance(item, Field)


def get_multi_value(key: str, data: dict, multiple: bool = False):
    return multiple and data.get('%s*' % key) or data.get(key)


def is_prop_key(name: str):
    return name not in reserved_keywords and re_prop_key.match(name)



import re
from ..css_abbreviation.parser import CSSValue, CSSProperty, FunctionCall
from ..css_abbreviation.tokenizer import tokens
from ..output_stream import OutputStream
from ..config import Config
from .color import color, frac
from .scope import CSSAbbreviationScope


def stringify(abbr: list, config: Config):
    out = OutputStream(config.options)
    fmt = config.options.get('output.format')

    if config.options.get('stylesheet.skipUnmatched'):
        # Filter out unmatched snippets
        abbr = [node for node in abbr if node.snippet is not None or node.important]

    for i, prop in enumerate(abbr):
        if fmt and i != 0:
            out.push_newline(True)
        css_property(prop, out, config)

    return out.value

def css_property(node: CSSProperty, out: OutputStream, config: Config):
    "Outputs given abbreviation node into output stream"
    is_json = config.options.get('stylesheet.json')

    if node.name:
        # It’s a CSS property
        name = to_camel_case(node.name) if is_json else node.name
        out.push_string(name + config.options.get('stylesheet.between'))

        if node.value:
            css_property_value(node, out, config)
        else:
            out.push_field(0, '')

        if is_json:
            # For CSS-in-JS, always finalize property with comma
            # NB: seems like `important` is not available in CSS-in-JS syntaxes
            out.push(',')
        else:
            output_important(node, out, True)
            out.push(config.options.get('stylesheet.after'))
    else:
        # It’s a regular snippet, output plain tokens without any additional formatting
        for css_val in node.value:
            for v in css_val.value:
                output_token(v, out, config)
        output_important(node, out, len(node.value) > 0)


def css_property_value(node: CSSProperty, out: OutputStream, config: Config):
    is_json = config.options.get('stylesheet.json')
    num = get_single_numeric(node) if is_json else None

    if num and (not num.unit or num.unit == 'px'):
        # For CSS-in-JS, if property contains single numeric value, output it
        # as JS number
        out.push(frac(num.value))
    else:
        quote = get_quote(config)
        if is_json: out.push(quote)
        for i, v in enumerate(node.value):
            if i != 0:
                out.push(', ')
            output_value(v, out, config)
        if is_json: out.push(quote)


def output_important(node: CSSProperty, out: OutputStream, separator=False):
    if node.important:
        if separator: out.push(' ')
        out.push('!important')


def output_value(value: CSSValue, out: OutputStream, config: Config):
    prev_end = -1
    for i, token in enumerate(value.value):
        # Handle edge case: a field is written close to previous token like this: `foo${bar}`.
        # We should not add delimiter here

        if i != 0 and (not isinstance(token, tokens.Field) or token.start != prev_end):
            out.push(' ')

        output_token(token, out, config)
        prev_end = token.end if hasattr(token, 'end') else -1

def output_token(token, out: OutputStream, config: Config):
    if isinstance(token, tokens.ColorValue):
        out.push(color(token, config.options.get('stylesheet.shortHex')))
    elif isinstance(token, tokens.Literal) or isinstance(token, tokens.CustomProperty):
        out.push_string(token.value)
    elif isinstance(token, tokens.NumberValue):
        out.push_string(frac(token.value, 4) + token.unit)
    elif isinstance(token, tokens.StringValue):
        quote = '"' if token.quote == 'double' else '\''
        out.push_string(''.join((quote, token.value, quote)))
    elif isinstance(token, tokens.Field):
        out.push_field(token.index, token.name)
    elif isinstance(token, FunctionCall):
        out.push(token.name + '(')
        for i, arg in enumerate(token.arguments):
            if i: out.push(', ')
            output_value(arg, out, config)
        out.push(')')


def get_single_numeric(node: CSSProperty):
    "If value of given property is a single numeric value, returns this token"
    if len(node.value) == 1:
        css_val = node.value[0]
        if len(css_val.value) == 1 and isinstance(css_val.value[0], tokens.NumberValue):
            return css_val.value[0]


def to_camel_case(text: str):
    "Converts kebab-case string to camelCase"
    return re.sub(r'\-(\w)', lambda m: m.group(1).upper(), text)

def get_quote(config: Config):
    return '"' if config.options.get('stylesheet.jsonDoubleQuotes') else '\''



__doc__ = "Utility methods for lists"

def some(fn: callable, items: list):
    "Tests whether at least one element in the array passes the test implemented by the provided function. It returns a `bool` value."
    if items:
        for item in items:
            if fn(item):
                return True
    return False


def find_index(fn: callable, items: list):
    "Returns the index of the first element in the array that satisfies the provided testing function. Otherwise, it returns -1, indicating that no element passed the test."
    for i, item in enumerate(items):
        if fn(item):
            return i
    return -1

def get_item(items: list, index: int):
    "Returns item by index from given list or `None` if index is out of range"
    return items[index] if 0 <= index < len(items) else None



import re
from .parser import TokenGroup, TokenElement, TokenAttribute, is_quote, is_bracket
from .tokenizer import tokens
from .stringify import stringify

re_url = re.compile(r'(https?:|ftp:|file:)?\/\/|(www|ftp)\.')
re_email = re.compile(r'[a-z0-9._%+-]+@[a-z0-9.-]+\.[a-z]{2,5}$', re.I)

class ConvertState:
    __slots__ = ('inserted', 'text', 'repeat_guard', 'repeaters', 'variables',
                 '_text_inserted', 'clean_text')

    def __init__(self, text: str = None, variables={}, max_repeat=None):
        self.inserted = False
        self.text = text

        if isinstance(text, list):
            self.clean_text = [l for l in text if l.strip()]
        else:
            self.clean_text = text

        self.repeat_guard = max_repeat if max_repeat is not None else 1000000
        self.variables = variables
        self.repeaters = []
        self._text_inserted = False

    def get_text(self, pos: int):
        self._text_inserted = True
        if isinstance(self.text, list):
            if pos is not None and pos >= 0 and pos < len(self.clean_text):
                return self.clean_text[pos].strip()

            value = self.text[pos] if pos is not None else '\n'.join(self.text)
        else:
            value = self.text or ''
        return value

    def get_variable(self, name: str):
        return self.variables.get(name) if self.variables else name


class Abbreviation:
    __slots__ = ('type', 'children')

    def __init__(self):
        self.type = 'Abbreviation'
        self.children = []

class AbbreviationNode:
    __slots__ = ('type', 'name', 'value', 'repeat', 'attributes', 'children', 'self_closing')

    def __init__(self, node: TokenElement, state: ConvertState):
        self.type = 'AbbreviationNode'
        self.name = stringify_name(node.name, state) if node.name else None
        self.value = stringify_value(node.value, state) if node.value else None
        self.attributes = None
        self.children = []
        self.repeat = clone_repeater(node.repeat) if node.repeat else None
        self.self_closing = node.self_close
        "Indicates current element is self-closing, e.g. should not contain closing pair"


class AbbreviationAttribute:
    __slots__ = ('name', 'value', 'value_type', 'boolean', 'implied', 'multiple')

    def __init__(self, name: str, value: list, value_type='raw', boolean=False, implied=False, multiple=False):
        self.name = name
        self.value = value
        self.value_type = value_type
        "Indicates type of value stored in `.value` property"

        self.boolean = boolean
        "Attribute is boolean (e.g.name equals value)"

        self.implied = implied
        "Attribute is implied (e.g.must be outputted only if contains non-null value)"

        self.multiple = multiple
        "Indicates that current attribute was repeated multiple times in a row"

    def copy(self):
        return AbbreviationAttribute(self.name, self.value, self.value_type, self.boolean, self.implied, self.multiple)


def convert(abbr: TokenGroup, params={}):
    "Converts given token-based abbreviation into simplified and unrolled node-based abbreviation"
    text = params.get('text')
    options = params.get('options') or {}
    state = ConvertState(text, params.get('variables'), params.get('max_repeat'))
    result = Abbreviation()
    result.children = convert_group(abbr, state)

    if text is not None and not state._text_inserted:
        # Text given but no implicitly repeated elements: insert it into deepest child
        deepest = deepest_node(result.children[-1])
        if deepest:
            tx = '\n'.join(text).strip() if isinstance(text, list) else text.strip() or ''
            insert_text(deepest, tx)
            if deepest.name == 'a' and options.get('markup.href'):
                # Automatically update value of `<a>` element if inserting URL or email
                insert_href(deepest, tx)


    return result


def convert_statement(node: TokenElement, state: ConvertState):
    result = []

    if node.repeat:
        # Node is repeated: we should create copies of given node
        # and supply context token with actual repeater state
        original = node.repeat
        repeat = clone_repeater(node.repeat)

        if repeat.implicit and isinstance(state.text, list):
            repeat.count = len(state.clean_text)
        else:
            repeat.count = repeat.count or 1

        state.repeaters.append(repeat)
        i = 0

        while i < repeat.count:
            repeat.value = i
            node.repeat = repeat
            items = convert_group(node, state) if is_group(node) else convert_element(node, state)

            if repeat.implicit and not state.inserted:
                # It’s an implicit repeater but no repeater placeholders found inside,
                # we should insert text into deepest node
                target = items[-1]
                deepest = deepest_node(target) if target else None
                if deepest:
                    insert_text(deepest, state.get_text(repeat.value))

            result += items

            # We should output at least one repeated item even if it’s reached
            # repeat limit
            state.repeat_guard -= 1
            if state.repeat_guard <= 0: break

            i += 1

        state.repeaters.pop()
        node.repeat = original

        if repeat.implicit: state.inserted = True
    else:
        result += convert_group(node, state) if is_group(node) else convert_element(node, state)

    return result


def convert_element(node: TokenElement, state: ConvertState):
    elem = AbbreviationNode(node, state)
    result = [elem]

    for child in node.elements:
        elem.children += convert_statement(child, state)

    if node.attributes:
        elem.attributes = [convert_attribute(attr, state) for attr in node.attributes]

    # In case if current node is a text-only snippet without fields, we should
    # put all children as siblings
    if not elem.name and elem.attributes is None and elem.value and not some(elem.value, is_field):
        result += elem.children
        elem.children = []

    return result


def convert_group(node: TokenGroup, state: ConvertState):
    result = []
    for child in node.elements:
        result += convert_statement(child, state)

    if node.repeat:
        result = attach_repeater(result, node.repeat)

    return result


def convert_attribute(node: TokenAttribute, state: ConvertState):
    attr = create_attribute(node, state)

    if node.value:
        tokens = node.value[:]

        if is_quote(tokens[0]):
            # It’s a quoted value: remove quotes from output but mark attribute
            # value as quoted
            quote = tokens.pop(0)

            if len(tokens) and tokens[-1].type == quote.type:
                tokens.pop()

            attr.value_type = 'singleQuote' if quote.single else 'doubleQuote'
        elif is_bracket(tokens[0], 'expression', True):
            # Value is expression: remove brackets but mark value type
            attr.value_type = 'expression'
            tokens.pop(0)

            if tokens and is_bracket(tokens[-1], 'expression', False):
                tokens.pop()

        attr.value = stringify_value(tokens, state)

    return attr


def create_attribute(node: TokenAttribute, state: ConvertState):
    name = stringify_name(node.name, state) if node.name else None
    value_type = 'expression' if node.expression else 'raw'
    boolean = False
    implied = False
    if name:
        if name[-1] == '.':
            boolean = True
            name = name[0:-1]
        if name[0] == '!':
            implied = True
            name = name[1:]
    return AbbreviationAttribute(name, None, value_type, boolean, implied, node.multiple)

def stringify_name(tokens: list, state: ConvertState):
    "Converts given token list to string"
    return ''.join([stringify(token, state) for token in tokens])


def stringify_value(token_list: list, state: ConvertState):
    "Converts given token list to value list"
    result = []
    accum = []

    for token in token_list:
        if is_field(token):
            # We should keep original fields in output since some editors has their
            # own syntax for field or doesn’t support fields at all so we should
            # capture actual field location in output stream
            if accum:
                result.append(''.join(accum))
                accum = []

            result.append(token)
        else:
            accum.append(stringify(token, state))


    if accum:
        result.append(''.join(accum))

    return result


def is_group(node):
    return isinstance(node, TokenGroup)


def is_field(token):
    return isinstance(token, tokens.Field) and token.index is not None


def deepest_node(node: AbbreviationNode):
    return deepest_node(node.children[-1]) if node.children else node


def insert_text(node: AbbreviationNode, text: str):
    if node.value:
        last_token = node.value[-1]
        if isinstance(last_token, str):
            node.value[-1] += text
        else:
            node.value.append(text)
    else:
        node.value = [text]


def insert_href(node: AbbreviationNode, text: str):
    href = None
    if re_url.match(text):
        href = text
        if not re.match(r'\w+:', href) and not href.startswith('//'):
            href = 'http://%s' % href
    elif re_email.match(text):
        href = 'mailto:%s' % text

    if href:
        href_attr = None
        if node.attributes:
            for attr in node.attributes:
                if attr.name == 'href':
                    href_attr = attr
                    break

        if not href_attr:
            href_attr = AbbreviationAttribute('href', href)
            if node.attributes:
                node.attributes.append(href_attr)
            else:
                node.attributes = [href_attr]
        elif not href_attr.value:
            href_attr.value = href


def attach_repeater(items: list, repeater: tokens.Repeater):
    for item in items:
        if not item.repeat:
            item.repeat = clone_repeater(repeater)

    return items


def clone_repeater(repeater: tokens.Repeater):
    return tokens.Repeater(repeater.count, repeater.value, repeater.implicit)

def some(items: list, fn: callable):
    for item in items:
        if fn(item): return True
    return False



from .snippets import markup_snippets, stylesheet_snippets, xsl_snippets, pug_snippets, variables

DEFAULT_SYNTAXES = { 'markup': 'html', 'stylesheet': 'css' }
"Default syntaxes for abbreviation types"

SYNTAXES = {
    'markup': ['html', 'xml', 'xsl', 'jsx', 'js', 'pug', 'slim', 'haml', 'vue', 'svelte'],
    'stylesheet': ['css', 'sass', 'scss', 'less', 'sss', 'stylus']
}
"List of all known syntaxes"

DEFAULT_OPTIONS = {
    'inlineElements': [
        'a', 'abbr', 'acronym', 'applet', 'b', 'basefont', 'bdo',
        'big', 'br', 'button', 'cite', 'code', 'del', 'dfn', 'em', 'font', 'i',
        'iframe', 'img', 'input', 'ins', 'kbd', 'label', 'map', 'object', 'q',
        's', 'samp', 'select', 'small', 'span', 'strike', 'strong', 'sub', 'sup',
        'textarea', 'tt', 'u', 'var'
    ],
    'output.indent': '\t',
    'output.baseIndent': '',
    'output.newline': '\n',
    'output.tagCase': '',
    'output.attributeCase': '',
    'output.attributeQuotes': 'double',
    'output.format': True,
    'output.formatLeafNode': False,
    'output.formatSkip': ['html'],
    'output.formatForce': ['body'],
    'output.inlineBreak': 3,
    'output.compactBoolean': False,
    'output.booleanAttributes': [
        'contenteditable', 'seamless', 'async', 'autofocus',
        'autoplay', 'checked', 'controls', 'defer', 'disabled', 'formnovalidate',
        'hidden', 'ismap', 'loop', 'multiple', 'muted', 'novalidate', 'readonly',
        'required', 'reversed', 'selected', 'typemustmatch'
    ],
    'output.reverseAttributes': False,
    'output.selfClosingStyle': 'html',
    'output.field': lambda index, placeholder, **kwargs: placeholder,
    'output.text': lambda text, **kwargs: text,

    'markup.href': True,

    'comment.enabled': False,
    'comment.trigger': ['id', 'class'],
    'comment.before': '',
    'comment.after': '\n<!-- /[#ID][.CLASS] -->',

    'bem.enabled': False,
    'bem.element': '__',
    'bem.modifier': '_',

    'jsx.enabled': False,

    'stylesheet.keywords': ['auto', 'inherit', 'unset', 'none'],
    'stylesheet.unitless': ['z-index', 'line-height', 'opacity', 'font-weight', 'zoom', 'flex', 'flex-grow', 'flex-shrink'],
    'stylesheet.shortHex': True,
    'stylesheet.between': ': ',
    'stylesheet.after': ';',
    'stylesheet.intUnit': 'px',
    'stylesheet.floatUnit': 'em',
    'stylesheet.unitAliases': { 'e': 'em', 'p': '%', 'x': 'ex', 'r': 'rem' },
    'stylesheet.json': False,
    'stylesheet.jsonDoubleQuotes': False,
    'stylesheet.fuzzySearchMinScore': 0,
    'stylesheet.skipUnmatched': True,
}

DEFAULT_CONFIG = {
    'type': 'markup',
    'syntax': 'html',
    'variables': variables,
    'snippets': {},
    'options': DEFAULT_OPTIONS
}

SYNTAX_CONFIG = {
    'markup': {
        'snippets': markup_snippets,
    },
    'xhtml': {
        'options': {
            'output.selfClosingStyle': 'xhtml'
        }
    },
    'xml': {
        'options': {
            'output.selfClosingStyle': 'xml'
        }
    },
    'xsl': {
        'snippets': xsl_snippets,
        'options': {
            'output.selfClosingStyle': 'xml'
        }
    },
    'jsx': {
        'options': {
            'jsx.enabled': True,
            'markup.attributes': {
                'class': 'className',
                'class*': 'styleName',
                'for': 'htmlFor'
            },
            'markup.valuePrefix': {
                'class*': 'styles'
            }
        }
    },
    'vue': {
        'options': {
            'markup.attributes': {
                'class*': ':class',
            }
        }
    },
    'svelte': {
        'options': {
            'jsx.enabled': True
        }
    },
    'pug': {
        'snippets': pug_snippets
    },

    'stylesheet': {
        'snippets': stylesheet_snippets
    },

    'sass': {
        'options': {
            'stylesheet.after': ''
        }
    },
    'stylus': {
        'options': {
            'stylesheet.between': ' ',
            'stylesheet.after': '',
        }
    }
}

class Config:
    __slots__ = ('type', 'syntax', 'variables', 'snippets', 'options', 'user_config', 'context', 'cache')

    def __init__(self, user_config={}, global_config={}):
        syntax_type = user_config.get('type', 'markup')
        syntax = user_config.get('syntax', DEFAULT_SYNTAXES.get(syntax_type, 'html'))

        self.type = syntax_type
        self.syntax = syntax
        self.user_config = user_config
        self.context = user_config.get('context')
        self.variables = merged_data(syntax_type, syntax, 'variables', user_config, global_config)
        self.snippets = merged_data(syntax_type, syntax, 'snippets', user_config, global_config)
        self.options = merged_data(syntax_type, syntax, 'options', user_config, global_config)
        self.cache = user_config.get('cache')

    def get(self, key: str):
        if key in dir(self):
            return self.__getattribute__(key)

        if key in self.user_config:
            return self.user_config.get(key)

        return None


def merged_data(syntax_type: str, syntax: str, key: str, user_config: dict, global_config: dict={}):
    empty = {}
    type_defaults = SYNTAX_CONFIG.get(syntax_type, empty)
    type_override = global_config.get(syntax_type, empty)
    syntax_defaults = SYNTAX_CONFIG.get(syntax, empty)
    syntax_override = global_config.get(syntax, empty)

    result = {}
    result.update(DEFAULT_CONFIG.get(key, empty))
    if key in type_defaults: result.update(type_defaults[key])
    if key in syntax_defaults: result.update(syntax_defaults[key])
    if key in type_override: result.update(type_override[key])
    if key in syntax_override: result.update(syntax_override[key])
    result.update(user_config.get(key, empty))

    return result



from ...abbreviation import Abbreviation, AbbreviationNode
from ...config import Config
from ...output_stream import OutputStream


class WalkState:
    __slots__ = ('current', 'parent', 'ancestors', 'config', 'out', 'field')

    def __init__(self, config: Config):
        self.current = None
        "Context node"

        self.parent = None
        "Immediate parent of currently iterated method"

        self.ancestors = []
        "List of all ancestors of context node"

        self.config = config
        "Current output config"

        self.out = OutputStream(config.options)
        "Output stream"

        self.field = 1
        "Current field index, used to output field marks for editor tabstops"

def walk(abbr: Abbreviation, visitor: callable, state: WalkState):
    def callback(ctx: AbbreviationNode, index: int, items: list):
        parent = state.parent
        current = state.current

        state.parent = current
        state.current = ctx
        visitor(ctx, index, items, state, walk_next)
        state.current = current
        state.parent = parent

    def walk_next(node, index, items):
        state.ancestors.append(state.current)
        callback(node, index, items)
        state.ancestors.pop()

    for index, child in enumerate(abbr.children):
        callback(child, index, abbr.children)



class TokenScanner:
    def __init__(self, tokens: list):
        self.tokens = tokens
        self.start = 0
        self.pos = 0
        self.size = len(tokens)

    def peek(self):
        return self.tokens[self.pos] if self.readable() else None

    def next(self):
        t = self.peek()
        self.pos += 1
        return t

    def slice(self, start: int=None, end: int=None):
        if start is None: start = self.start
        if end is None: end = self.pos
        return self.tokens[start:end]

    def readable(self):
        return self.pos < self.size

    def consume(self, test: callable):
        token = self.peek()
        if token and test(token):
            self.pos += 1
            return True

        return False

    def consume_while(self, test: callable):
        start = self.pos
        while self.consume(test):
            pass
        return self.pos != start

    def error(self, message: str, token=None):
        if token is None:
            token = self.peek()
        pos = None
        if token and token.start is not None:
            pos = token.start
            message += ' at %d' % pos
        return TokenScannerException(message, pos)

class TokenScannerException(Exception):
    def __init__(self, message: str, pos: int):
        super(TokenScannerException, self).__init__(message)
        self.message = message
        self.pos = pos



from ..scanner import Scanner
from ..scanner_utils import is_space, eat_quoted
from .utils import ElementType, Chars, consume_array, is_terminator, consume_section, ident
from .attributes import attributes, attribute_name, attribute_value, get_attribute_value

cdata_open = '<![CDATA['
cdata_close = ']]>'
comment_open = '<!--'
comment_close = '-->'
pi_start = '<?'
pi_end = '?>'

def scan(source: str, callback: callable, special: dict=None):
    """
    Performs fast scan of given source code: for each tag found it invokes callback
    with tag name, its type (open, close, self-close) and range in original source.
    Unlike regular scanner, fast scanner doesn’t provide info about attributes to
    reduce object allocations hence increase performance.
    If `callback` returns `false`, scanner stops parsing.
    :param special List of “special” HTML tags which should be ignored. Most likely
    it’s a "script" and "style" tags.
    """
    scanner = Scanner(source)
    found = False

    while not scanner.eof():
        if cdata(scanner) or comment(scanner) or processing_instruction(scanner):
            continue

        start = scanner.pos
        if scanner.eat(Chars.LeftAngle):
            # Maybe a tag name?
            elem_type = ElementType.Close if scanner.eat(Chars.Slash) else ElementType.Open
            name_start = scanner.pos

            if ident(scanner):
                # Consumed tag name
                name_end = scanner.pos
                if elem_type != ElementType.Close:
                    skip_attributes(scanner)
                    scanner.eat_while(is_space)
                    if scanner.eat(Chars.Slash):
                        elem_type = ElementType.SelfClose

                if scanner.eat(Chars.RightAngle):
                    # Tag properly closed
                    name = scanner.substring(name_start, name_end)
                    if callback(name, elem_type, start, scanner.pos) is False:
                        break

                    if elem_type == ElementType.Open and special and is_special(special, name, source, start, scanner.pos):
                        # Found opening tag of special element: we should skip
                        # scanner contents until we find closing tag
                        found = False
                        while not scanner.eof():
                            if consume_closing(scanner, name):
                                found = True
                                break

                            scanner.pos += 1

                        if found and callback(name, ElementType.Close, scanner.start, scanner.pos) is False:
                            break
        else:
            scanner.pos += 1


def skip_attributes(scanner: Scanner):
    "Skips attributes in current tag context"
    while not scanner.eof():
        scanner.eat_while(is_space)
        if attribute_name(scanner):
            if scanner.eat(Chars.Equals):
                attribute_value(scanner)
        elif is_terminator(scanner.peek()):
            break
        else:
            scanner.pos += 1

def consume_closing(scanner: Scanner, name: str):
    "Consumes closing tag with given name from scanner"
    start = scanner.pos
    if scanner.eat(Chars.LeftAngle) and \
        scanner.eat(Chars.Slash) and \
        consume_array(scanner, name) and \
        scanner.eat(Chars.RightAngle):
        scanner.start = start
        return True

    scanner.pos = start
    return False


def cdata(scanner: Scanner):
    "Consumes CDATA from given scanner"
    global cdata_open, cdata_close
    return consume_section(scanner, cdata_open, cdata_close, True)


def comment(scanner: Scanner):
    "Consumes comments from given scanner"
    global comment_open, comment_close
    return consume_section(scanner, comment_open, comment_close, True)


def processing_instruction(scanner: Scanner):
    "Consumes processing instruction from given scanner"
    global pi_start, pi_end
    if consume_array(scanner, pi_start):
        while not scanner.eof():
            if consume_array(scanner, pi_end):
                break

            if not eat_quoted(scanner): scanner.pos += 1

        return True

    return False


def is_special(special: dict, name: str, source: str, start: int, end: int):
    "Check if given tag name should be considered as special"
    if name in special:
        type_values = special[name]
        if not isinstance(type_values, list):
            return True

        attrs = attributes(source[start + len(name) + 1:end - 1])
        value = get_attribute_value(attrs, 'type') or ''
        return value in type_values

    return False



import unittest
import sys
import re

sys.path.append('../')

from emmet import expand

def word_count(text: str):
    return len(text.split(' '))


class TestLoremIpsum(unittest.TestCase):
    def test_single(self):
        output = expand('lorem')
        self.assertTrue(re.match(r'^Lorem,?\sipsum', output))
        self.assertTrue(word_count(output) > 20)

        output = expand('lorem5')
        self.assertTrue(re.match(r'^Lorem,?\sipsum', output))
        self.assertEqual(word_count(output), 5)

        output = expand('lorem5-10')
        self.assertTrue(re.match(r'^Lorem,?\sipsum', output))
        self.assertTrue(5 <= word_count(output) <= 10)

        output = expand('loremru4')
        self.assertTrue(re.match(r'^Далеко-далеко,?\sза,?\sсловесными', output))
        self.assertEqual(word_count(output), 4)

        output = expand('p>lorem')
        self.assertTrue(re.match(r'^<p>Lorem,?\sipsum', output))

        # https://github.com/emmetio/expand-abbreviation/issues/24
        output = expand('(p)lorem2')
        self.assertTrue(re.match(r'^<p><\/p>\nLorem,?\sipsum', output))

        output = expand('p(lorem10)')
        self.assertTrue(re.match(r'^<p><\/p>\nLorem,?\sipsum', output))

    def test_multiple(self):
        output = expand('lorem6*3')
        lines = output.splitlines()
        self.assertTrue(re.match(r'^Lorem,?\sipsum', output))
        self.assertEqual(len(lines), 3)

        output = expand('lorem6*2')
        lines = output.splitlines()
        self.assertTrue(re.match(r'^Lorem,?\sipsum', output))
        self.assertEqual(len(lines), 2)

        output = expand('p*3>lorem')
        lines = output.splitlines()
        self.assertTrue(re.match(r'^<p>Lorem,?\sipsum', lines[0]))
        self.assertFalse(re.match(r'^<p>Lorem,?\sipsum', lines[1]))

        output = expand('ul>lorem5*3', { 'options': { 'output.indent': '' } })
        lines = output.splitlines()
        self.assertEqual(len(lines), 5)
        self.assertTrue(re.match(r'^^<li>Lorem,?\sipsum', lines[1]))
        self.assertFalse(re.match(r'^^<li>Lorem,?\sipsum', lines[2]))



from ..abbreviation import Abbreviation, AbbreviationNode

def walk(node: AbbreviationNode, fn: callable, state=None):
    """
    Walks over each child node of given markup abbreviation AST node (not including
    given one) and invokes `fn` on each node.
    The `fn` callback accepts context node, list of ancestor nodes and optional
    state object
    """
    ancestors = [node]
    def callback(ctx: AbbreviationNode):
        fn(ctx, ancestors, state)
        ancestors.append(ctx)
        for child in ctx.children:
            callback(child)
        ancestors.pop()

    for child in node.children:
        callback(child)

def find_deepest(node: AbbreviationNode):
    "Finds node which is the deepest for in current node or node itself."
    parent = None
    while node.children:
        parent = node
        node = node.children[-1]

    return (parent, node)


def find(node: AbbreviationNode, callback: callable):
    "Finds first child node that matches given `callback`"
    for child in node.children:
        if callback(child):
            return child

        result = find(child, callback)
        if result:
            return result



import re

from .walk import walk, WalkState
from .utils import push_tokens, caret, split_by_lines, is_snippet, should_output_attribute
from ...abbreviation import Abbreviation, AbbreviationNode, AbbreviationAttribute
from ...abbreviation.tokenizer.tokens import Field
from ...config import Config
from ...output_stream import attr_name, attr_quote, is_boolean_attribute


__doc__ =  "Utility methods for working with indent-based markup languages like HAML, Slim, Pug etc."


class IndentWalkState(WalkState):
    __slots__ = ('options')


def indent_format(abbr: Abbreviation, config: Config, options={}):
    state = IndentWalkState(config)
    state.options = options
    walk(abbr, element, state)
    return state.out.value


def element(node: AbbreviationNode, index: int, items: list, state: IndentWalkState, walk_next: callable):
    out = state.out
    options = state.options
    primary, secondary = collect_attributes(node)

    # Pick offset level for current node
    level = 1 if state.parent else 0
    out.level += level

    # Do not indent top-level elements
    if should_format(node, index, items, state):
        out.push_newline(True)

    if node.name and (node.name != 'div' or not primary):
        s = '%s%s%s' % (options.get('beforeName', ''), node.name, options.get('afterName', ''))
        out.push_string(s)

    push_primary_attributes(primary, state)
    push_secondary_attributes(list(filter(should_output_attribute, secondary)), state)

    if node.self_closing and not node.value and not node.children:
        if state.options['selfClose']:
            out.push_string(state.options['selfClose'])
    else:
        push_value(node, state)
        for index, child in enumerate(node.children):
            walk_next(child, index, node.children)

    out.level -= level


def collect_attributes(node: AbbreviationNode):
    """
    From given node, collects all attributes as `primary` (id, class) and
    `secondary` (all the rest) lists. In most indent-based syntaxes, primary attribute
    has special syntax
    """
    primary = []
    secondary = []

    if node.attributes:
        for attr in node.attributes:
            if is_primary_attribute(attr):
                primary.append(attr)
            else:
                secondary.append(attr)

    return (primary, secondary)


def push_primary_attributes(attrs: list, state: WalkState):
    "Outputs given attributes as primary into output stream"
    for attr in attrs:
        if attr.value is not None:
            if attr.name == 'class':
                state.out.push_string('.')
                # All whitespace characters must be replaced with dots in class names
                tokens = [re.sub(r'\s+', '.', t) if isinstance(t, str) else t for t in attr.value]
                push_tokens(tokens, state)
            else:
                # ID attribute
                state.out.push_string('#')
                push_tokens(attr.value, state)


def push_secondary_attributes(attrs: list, state: IndentWalkState):
    if attrs:
        out = state.out
        config = state.config
        options = state.options

        before = options.get('beforeAttribute')
        after = options.get('afterAttribute')
        glue = options.get('glueAttribute')

        if before: out.push_string(before)

        for i, attr in enumerate(attrs):
            out.push_string(attr_name(attr.name or '', config))
            if is_boolean_attribute(attr, config) and not attr.value:
                if not config.options.get('output.compactBoolean') and options.get('booleanValue'):
                    out.push_string('=%s' % options.get('booleanValue'))
            else:
                out.push_string('=%s' % attr_quote(attr, config, True))
                push_tokens(attr.value or caret, state)
                out.push_string(attr_quote(attr, config))

            if i != len(attrs) - 1 and glue:
                out.push_string(glue)

        if after: out.push_string(after)


def push_value(node: AbbreviationNode, state: IndentWalkState):
    # We should either output value or add caret but for leaf nodes only (no children)
    if not node.value and node.children:
        return

    value = node.value or caret
    lines = split_by_lines(value)
    out = state.out
    options = state.options

    if len(lines) == 1:
        if node.name or node.attributes:
            out.push(' ')
        push_tokens(value, state)
    else:
        # We should format multi-line value with terminating `|` character
        # and same line length
        line_lengths = []
        max_length = 0
        before = options.get('beforeTextLine')
        after = options.get('afterTextLine')

        # Calculate lengths of all lines and max line length
        for line in lines:
            l = value_length(line)
            line_lengths.append(l)
            if l > max_length:
                max_length = l

        # Output each line, padded to max length
        out.level += 1
        for i, line in enumerate(lines):
            out.push_newline(True)
            if before:
                out.push(before)
            push_tokens(line, state)
            if after:
                out.push(' ' * (max_length - line_lengths[i]))
                out.push(after)

        out.level -= 1

def is_primary_attribute(attr: AbbreviationAttribute):
    return attr.name == 'class' or attr.name == 'id'


def value_length(tokens: list):
    "Calculates string length from given tokens"
    l = 0

    for token in tokens:
        l += len(token) if isinstance(token, str) else len(token.name)

    return l

def should_format(node: AbbreviationNode, index: int, items: list, state: WalkState):
    # Do not format first top-level element or snippets
    if not state.parent and index == 0:
        return False

    return not is_snippet(node)



from .tokenizer import tokenize
from .convert import convert, Abbreviation, AbbreviationAttribute, AbbreviationNode
from .parser import parse as parser
from ..scanner import ScannerException

def parse(abbr: str, options={}):
    try:
        tokens = tokenize(abbr) if isinstance(abbr, str) else abbr
        return convert(parser(tokens, options), options)
    except ScannerException as err:
        if isinstance(abbr, str):
            err.message += '\n%s\n%s^' % (abbr, '-' * err.pos)

        raise err



from ..scanner_utils import is_space, is_number
from .parser import Operator, is_sign, is_operator

class BackwardScanner:
    __slots__ = ('text', 'pos')

    def __init__(self, text: str, pos=0):
        self.text = text
        self.pos = pos

    def prev(self):
        return self.text[self.pos - 1] if self.pos else ''

    def cur(self):
        return self.text[self.pos] if self.pos < len(self.text) else ''


def extract(text: str, pos=None, options: dict=None) -> tuple:
    """
    Extracts math expression from given text at specified position.
    Expression is extracted in backward direction.

    Options:
    :lookAhead bool
    Allow capturing extra expression characters right after start position.
    Useful for extracting expressions from text editor source which inserts
    paired characters like `(` and `)` to properly extract expression past
    caret position

    :whitespace: bool
    Allow whitespace in extracted expressions
    """
    if pos is None:
        pos = len(text)

    opt = { 'lookAhead': True, 'whitespace': True }
    if options: opt.update(options)

    scanner = BackwardScanner(text, pos)

    if opt['lookAhead'] and scanner.cur() == Operator.RightParenthesis:
        # Basically, we should consume right parenthesis only with optional whitespace
        scanner.pos += 1
        l = len(text)
        while scanner.pos < l:
            ch = scanner.cur()
            if ch != Operator.RightParenthesis and not (opt['whitespace'] and is_space(ch)):
                break
            scanner.pos += 1

    end = scanner.pos
    braces = 0
    while scanner.pos >= 0:
        if number(scanner):
            continue

        ch = scanner.prev()
        if ch == Operator.RightParenthesis:
            braces += 1
        elif ch == Operator.LeftParenthesis:
            if not braces:
                break
            braces -= 1
        elif not ((opt['whitespace'] and is_space(ch)) or is_sign(ch) or is_operator(ch)):
            break

        scanner.pos -= 1

    if scanner.pos != end and not braces:
        # Trim whitespace
        while is_space(scanner.cur()):
            scanner.pos += 1

        return (scanner.pos, end)


def number(scanner: BackwardScanner):
    if is_number(scanner.prev()):
        scanner.pos -= 1
        dot = False

        while scanner.pos >= 0:
            ch = scanner.prev()
            if ch == '.':
                if dot:
                    # Decimal delimiter already consumed, abort
                    break
                dot = True
            elif not is_number(ch):
                break
            scanner.pos -= 1

        return True

    return False



import re
from ..css_abbreviation.tokenizer.tokens import ColorValue

def color(token: ColorValue, short_hex=False):
    if not token.r and not token.g and not token.b and not token.a:
        return 'transparent'
    if token.a == 1:
        return as_hex(token, short_hex)

    return as_rgb(token)

def as_hex(token: ColorValue, short=False):
    """
    Output given color as hex value
    :param short Produce short value (e.g. #fff instead of #ffffff), if possible
    """
    if short and is_short_hex(token.r) and is_short_hex(token.g) and is_short_hex(token.b):
        fn = to_short_hex
    else:
        fn = to_hex

    return '#%s%s%s' % (fn(token.r), fn(token.g), fn(token.b))

def as_rgb(token: ColorValue):
    values = [str(token.r), str(token.g), str(token.b)]
    prefix = 'rgb'
    if token.a != 1:
        prefix = 'rgba'
        values.append(frac(token.a, 8))

    return '%s(%s)' % (prefix, ', '.join(values))

def frac(num: int, digits=4):
    str_num = ('%.' + str(digits) + 'f') % num
    return re.sub(r'\.?0+$', '', str_num)

def is_short_hex(num: int):
    return not (num % 17)

def to_short_hex(num: int):
    return format(num >> 4, 'x')

def to_hex(num: int):
    return format(num, 'x').ljust(2, '0')



nbsp = chr(160)

def is_number(ch: str):
    "Check if given code is a number"
    return ch.isdecimal()

def is_alpha(ch: str):
    "Check if given character code is alpha code (letter through A to Z)"
    return 'a' <= ch <= 'z' or 'A' <= ch <= 'Z'

def is_alpha_numeric(ch: str):
    "Check if given character code is alpha-numeric (letter through A to Z or number)"
    return is_number(ch) or is_alpha(ch)

def is_alpha_numeric_word(ch: str):
    return is_number(ch) or is_alpha_word(ch)

def is_alpha_word(ch: str):
    return ch == '_' or is_alpha(ch)

def is_white_space(ch: str):
    "Check if given character code is a white-space character: a space character without line breaks"
    return ch == ' ' or ch == '\t' or ch == nbsp

def is_space(ch: str):
    "Check if given character code is a space character"
    return is_white_space(ch) or ch == '\n' or ch == '\r'

def is_quote(ch: str):
    "Check if given character code is a quote character"
    return ch == '"' or ch == "'"

def eat_quoted(scanner, options={}):
    """
    Consumes 'single' or "double"-quoted string from given string, if possible
    Returns `True` if quoted string was consumed. The contents of quoted string
    will be available as `scanner.current()`
    """
    options = create_options(options)
    start = scanner.pos
    quote = scanner.peek()

    if scanner.eat(is_quote):
        while not scanner.eof():
            if scanner.eat(quote):
                scanner.start = start
                return True
            scanner.eat(options['escape'])
            scanner.pos += 1

        # If we’re here then stream wasn’t properly consumed.
        # Revert stream and decide what to do
        scanner.pos = start

        if options['throws']:
            raise scanner.error('Unable to consume quoted string')

    return False

def eat_pair(scanner, open_ch: str, close_ch: str, options={}):
    options = create_options(options)
    start = scanner.pos

    if scanner.eat(open_ch):
        stack = 1

        while not scanner.eof():
            if eat_quoted(scanner, options): continue

            ch = scanner.next()
            if ch == open_ch:
                stack += 1
            elif ch == close_ch:
                stack -= 1
                if not stack:
                    scanner.start = start
                    return True
            elif ch == options['escape']:
                scanner.pos += 1

        # If we’re here then paired character can’t be consumed
        scanner.pos = start

        if options['throws']:
            raise scanner.error('Unable to find matching pair for %s' % open_ch)

    return False

def create_options(opt={}):
    options = {
        'escape': '\\',
        'throws': False
    }

    options.update(opt)
    return options



from .css import snippets as raw_stylesheet_snippets
from .html import snippets as raw_markup_snippets
from .xsl import snippets as raw_xsl_snippets
from .pug import snippets as raw_pug_snippets
from .variables import variables

def parse_snippets(snippets: dict):
    """
    Parses raw snippets definitions with possibly multiple keys into a plan
    snippet map
    """
    result = {}
    for k in snippets.keys():
        for name in k.split('|'):
            result[name] = snippets[k]

    return result

# Parse raw snippets which may contains multiple entries in key into one-to-one dict
markup_snippets = parse_snippets(raw_markup_snippets)
stylesheet_snippets = parse_snippets(raw_stylesheet_snippets)
xsl_snippets = parse_snippets(raw_xsl_snippets)
pug_snippets = parse_snippets(raw_pug_snippets)



import unittest
import sys

sys.path.append('../')

from emmet import expand_stylesheet, parse_stylesheet_snippets, Config
from emmet.stylesheet import CSSAbbreviationScope
from emmet.stylesheet.score import calculate_score as score

def field(index: int, placeholder: str, **kwargs):
    if placeholder:
        return '${%d:%s}' % (index, placeholder)
    return '${%d}' % index

default_config = Config({
    'type': 'stylesheet',
    'options': {
        'output.field': field
    },
    'snippets': {
        'mten': 'margin: 10px;',
        'fsz': 'font-size',
        'gt': 'grid-template: repeat(2,auto) / repeat(auto-fit, minmax(250px, 1fr))',
        'bxsh': 'box-shadow: var(--bxsh-${1})'
    },
    'cache': {}
})

def expand(abbr: str, config=default_config):
    return expand_stylesheet(abbr, config)


def pick(abbr: str, items: list):
    items = map(lambda item: { 'item': item, 'score': score(abbr, item, True) }, items)
    items = list(filter(lambda item: item['score'] > 0, items))
    items.sort(key=lambda item: item['score'], reverse=True)

    if items:
        return items[0]['item']


class TestStylesheetScoring(unittest.TestCase):
    def test_compare_scores(self):
        self.assertEqual(score('aaa', 'aaa'), 1)
        self.assertEqual(score('baa', 'aaa'), 0)

        self.assertFalse(score('b', 'aaa'))
        self.assertTrue(score('a', 'aaa'))
        self.assertTrue(score('a', 'abc'))
        self.assertTrue(score('ac', 'abc'))
        self.assertTrue(score('a', 'aaa') < score('aa', 'aaa'))
        self.assertTrue(score('ab', 'abc') > score('ab', 'acb'))

        # acronym bonus
        self.assertTrue(score('ab', 'a-b') > score('ab', 'acb'))

    def test_pick_padding_or_position(self):
        items = ['p', 'pb', 'pl', 'pos', 'pa', 'oa', 'soa', 'pr', 'pt']

        self.assertEqual(pick('p', items), 'p')
        self.assertEqual(pick('poa', items), 'pos')


class TestStylesheetAbbreviations(unittest.TestCase):
    def test_keyword(self):
        self.assertEqual(expand('bd1-s'), 'border: 1px solid;')
        self.assertEqual(expand('dib'), 'display: inline-block;')
        self.assertEqual(expand('bxsz'), 'box-sizing: ${1:border-box};')
        self.assertEqual(expand('bxz'), 'box-sizing: ${1:border-box};')
        self.assertEqual(expand('bxzc'), 'box-sizing: content-box;')
        self.assertEqual(expand('fl'), 'float: ${1:left};')
        self.assertEqual(expand('fll'), 'float: left;')

        self.assertEqual(expand('pos'), 'position: ${1:relative};')
        self.assertEqual(expand('poa'), 'position: absolute;')
        self.assertEqual(expand('por'), 'position: relative;')
        self.assertEqual(expand('pof'), 'position: fixed;')
        self.assertEqual(expand('pos-a'), 'position: absolute;')

        self.assertEqual(expand('m'), 'margin: ${0};')
        self.assertEqual(expand('m0'), 'margin: 0;')

        # use `auto` as global keyword
        self.assertEqual(expand('m0-a'), 'margin: 0 auto;')
        self.assertEqual(expand('m-a'), 'margin: auto;')

        self.assertEqual(expand('bg'), 'background: ${1:#000};')

        self.assertEqual(expand('bd'), 'border: ${1:1px} ${2:solid} ${3:#000};')
        self.assertEqual(expand('bd0-s#fc0'), 'border: 0 solid #fc0;')
        self.assertEqual(expand('bd0-dd#fc0'), 'border: 0 dot-dash #fc0;')
        self.assertEqual(expand('bd0-h#fc0'), 'border: 0 hidden #fc0;')

        self.assertEqual(expand('trf-trs'), 'transform: translate(${1:x}, ${2:y});')

        # https://github.com/emmetio/emmet/issues/647
        self.assertEqual(expand('gtc'), 'grid-template-columns: repeat(${0});')
        self.assertEqual(expand('gtr'), 'grid-template-rows: repeat(${0});')

        self.assertEqual(expand('lis:n'), 'list-style: none;')
        self.assertEqual(expand('list:n'), 'list-style-type: none;')
        self.assertEqual(expand('bdt:n'), 'border-top: none;')
        self.assertEqual(expand('bgi:n'), 'background-image: none;')
        self.assertEqual(expand('q:n'), 'quotes: none;')

        # Custom properties
        # https://github.com/emmetio/emmet/issues/692
        self.assertEqual(expand('bxsh'), 'box-shadow: var(--bxsh-${1});')

    def test_numeric(self):
        self.assertEqual(expand('p0'), 'padding: 0;', 'No unit for 0')
        self.assertEqual(expand('p10'), 'padding: 10px;', '`px` unit for integers')
        self.assertEqual(expand('p.4'), 'padding: 0.4em;', '`em` for floats')
        self.assertEqual(expand('fz10'), 'font-size: 10px;', '`px` for integers')
        self.assertEqual(expand('fz1.'), 'font-size: 1em;', '`em` for explicit float')
        self.assertEqual(expand('p10p'), 'padding: 10%;', 'unit alias')
        self.assertEqual(expand('z10'), 'z-index: 10;', 'Unitless property')
        self.assertEqual(expand('p10r'), 'padding: 10rem;', 'unit alias')
        self.assertEqual(expand('mten'), 'margin: 10px;', 'Ignore terminating `;` in snippet')

        # https://github.com/microsoft/vscode/issues/59951
        self.assertEqual(expand('fz'), 'font-size: ${0};')
        self.assertEqual(expand('fz12'), 'font-size: 12px;')
        self.assertEqual(expand('fsz'), 'font-size: ${0};')
        self.assertEqual(expand('fsz12'), 'font-size: 12px;')
        self.assertEqual(expand('fs'), 'font-style: ${1:italic};')

        # https://github.com/emmetio/emmet/issues/558
        self.assertEqual(expand('us'), 'user-select: none;')

        # https://github.com/emmetio/emmet/issues/558
        self.assertEqual(expand('us'), 'user-select: none;');

        # https://github.com/microsoft/vscode/issues/105697
        self.assertEqual(expand('opa1'), 'opacity: 1;', 'Unitless property');
        self.assertEqual(expand('opa.1'), 'opacity: 0.1;', 'Unitless property');
        self.assertEqual(expand('opa.a'), 'opacity: .a;', 'Unitless property');

    def test_numeric_with_format(self):
        config = Config({
            'type': 'stylesheet',
            'options': {
                'stylesheet.intUnit': 'pt',
                'stylesheet.floatUnit': 'vh',
                'stylesheet.unitAliases': {
                    'e': 'em',
                    'p': '%',
                    'x': 'ex',
                    'r': ' / @rem'
                }
            }
        })
        self.assertEqual(expand('p0', config), 'padding: 0;', 'No unit for 0')
        self.assertEqual(expand('p10', config), 'padding: 10pt;', '`pt` unit for integers')
        self.assertEqual(expand('p.4', config), 'padding: 0.4vh;', '`vh` for floats')
        self.assertEqual(expand('p10p', config), 'padding: 10%;', 'unit alias')
        self.assertEqual(expand('z10', config), 'z-index: 10;', 'Unitless property')
        self.assertEqual(expand('p10r', config), 'padding: 10 / @rem;', 'unit alias')

    def test_important(self):
        self.assertEqual(expand('!'), '!important')
        self.assertEqual(expand('p!'), 'padding: ${0} !important;')
        self.assertEqual(expand('p0!'), 'padding: 0 !important;')

    def test_color(self):
        self.assertEqual(expand('c'), 'color: ${1:#000};')
        self.assertEqual(expand('c#'), 'color: #000;')
        self.assertEqual(expand('c#f.5'), 'color: rgba(255, 255, 255, 0.5);')
        self.assertEqual(expand('c#f.5!'), 'color: rgba(255, 255, 255, 0.5) !important;')
        self.assertEqual(expand('bgc'), 'background-color: #${1:fff};')
        self.assertEqual(expand('bgc#f0'), 'background-color: #f0f0f0;')
        self.assertEqual(expand('bgc#f1'), 'background-color: #f1f1f1;')

    def test_snippets(self):
        self.assertEqual(expand('@k'), '@keyframes ${1:identifier} {\n\t${2}\n}')
        self.assertEqual(expand('@'), '@media ${1:screen} {\n\t${0}\n}')

        # Insert value into snippet fields
        self.assertEqual(expand('@k-name'), '@keyframes name {\n\t${2}\n}')
        self.assertEqual(expand('@k-name10'), '@keyframes name {\n\t10\n}')
        self.assertEqual(expand('gt'), 'grid-template: repeat(2, auto) / repeat(auto-fit, minmax(250px, 1fr));')

    def test_multiple_properties(self):
        self.assertEqual(expand('p10+m10-20'), 'padding: 10px;\nmargin: 10px 20px;')
        self.assertEqual(expand('p+bd'), 'padding: ${0};\nborder: ${1:1px} ${2:solid} ${3:#000};')

    def test_functions(self):
        self.assertEqual(expand('trf-s(2)'), 'transform: scale(2, ${2:y});')
        self.assertEqual(expand('trf-s(2, 3)'), 'transform: scale(2, 3);')

    def test_case_insensitive_match(self):
        self.assertEqual(expand('trf:rx'), 'transform: rotateX(${1:angle});')

    def test_gradient_resolver(self):
        self.assertEqual(expand('lg'), 'background-image: linear-gradient(${0});')
        self.assertEqual(expand('lg(to right, #0, #f00.5)'), 'background-image: linear-gradient(to right, #000, rgba(255, 0, 0, 0.5));')

    # This example is useless: it’s unexpected to receive `align-self: unset`
    # for `auto` snippet
    # def test_min_score(self):
    #     self.assertEqual(expand('auto', Config({
    #         'type': 'stylesheet',
    #         'options': { 'stylesheet.fuzzySearchMinScore': 0 }
    #     })), 'align-self: unset;')
    #     self.assertEqual(expand('auto', Config({
    #         'type': 'stylesheet',
    #         'options': { 'stylesheet.fuzzySearchMinScore': 0.3 }
    #     })), 'auto: ;')

    def test_css_in_js(self):
        config = Config({
            'type': 'stylesheet',
            'options': {
                'stylesheet.json': True,
                'stylesheet.between': ': '
            }
        })

        self.assertEqual(expand('p10+mt10-20', config), 'padding: 10,\nmarginTop: \'10px 20px\',')
        self.assertEqual(expand('bgc', config), 'backgroundColor: \'#fff\',')

    def test_resolve_context_value(self):
        config = Config({
            'type': 'stylesheet',
            'context': { 'name': 'align-content' }
        })

        self.assertEqual(expand('s', config), 'start')
        self.assertEqual(expand('a', config), 'auto')

    def test_limit_snippets_by_scope(self):
        section_scope = Config({
            'type': 'stylesheet',
            'context': { 'name': CSSAbbreviationScope.Section },
            'snippets': {
                'mten': 'margin: 10px;',
                'fsz': 'font-size',
                'myCenterAwesome': 'body {\n\tdisplay: grid;\n}'
            }
        })
        property_scope = Config({
            'type': 'stylesheet',
            'context': { 'name': CSSAbbreviationScope.Property },
            'snippets': {
                'mten': 'margin: 10px;',
                'fsz': 'font-size',
                'myCenterAwesome': 'body {\n\tdisplay: grid;\n}'
            }
        })

        self.assertEqual(expand('m', section_scope), 'body {\n\tdisplay: grid;\n}')
        self.assertEqual(expand('b', section_scope), '')
        self.assertEqual(expand('m', property_scope), 'margin: ;')



class Chars:
    CurlyBracketOpen = '{'
    CurlyBracketClose = '}'
    Escape = '\\'
    Equals = '='
    SquareBracketOpen = '['
    SquareBracketClose = ']'
    Asterisk = '*'
    Hash = '#'
    Dollar = '$'
    Dash = '-'
    Dot = '.'
    Slash = '/'
    Colon = ':'
    Excl = '!'
    At = '@'
    Underscore = '_'
    RoundBracketOpen = '('
    RoundBracketClose = ')'
    Sibling = '+'
    Child = '>'
    Climb = '^'
    SingleQuote = "'"
    DoubleQuote = '"'

def escaped(scanner):
    if scanner.eat(Chars.Escape):
        scanner.start = scanner.pos
        if not scanner.eof():
            scanner.pos += 1

        return True

    return False



import re
from .reader import BackwardScanner
from .is_html import is_html as is_at_html_tag, is_quote
from .brackets import Brackets, BRACE_PAIRS
from ..scanner_utils import is_alpha, is_number

class ExtractedAbbreviation:
    __slots__ = ('abbreviation', 'location', 'start', 'end')

    def __init__(self, abbreviation: str, location: int, start: int, end: int):
        self.abbreviation = abbreviation
        "Extracted abbreviation"

        self.location = location
        "Location of abbreviation in input string"

        self.start = start
        "Start location of matched abbreviation, including prefix"

        self.end = end
        "End location of extracted abbreviation"

    def __eq__(self, other):
        if isinstance(other, ExtractedAbbreviation):
            return self.abbreviation == other.abbreviation and \
                self.location == other.location and \
                self.start == other.start and \
                self.end == other.end

        raise NotImplementedError

    def __repr__(self):
        return repr({
            'abbreviation': self.abbreviation,
            'location': self.location,
            'start': self.start,
            'end': self.end,
        })

SPECIAL_CHARS = '#.*:$-_!@%^+>/'

def extract_abbreviation(line: str, pos: int=None, options={}) -> ExtractedAbbreviation:
    """
    Extracts abbreviation from given line of source code.
    Options:

    Parser options:

    lookAhead: bool
    Allow parser to look ahead of `pos` index for searching of missing
    abbreviation parts. Most editors automatically inserts closing braces for
    `[`, `{` and `(`, which will most likely be right after current caret position.
    So in order to properly expand abbreviation, user must explicitly move
    caret right after auto-inserted braces. With this option enabled, parser
    will search for closing braces right after `pos`. Default is `true`

    type: 'markup' | 'stylesheet'
    Type of context syntax of expanded abbreviation.
    In 'stylesheet' syntax, brackets `[]` and `{}` are not supported thus
    not extracted.

    prefix: str
    A string that should precede abbreviation in order to make it successfully
    extracted. If given, the abbreviation will be extracted from the nearest
    `prefix` occurrence.
    """
    if pos is None:
        pos = len(line)
    opt = create_options(options)
    # make sure `pos` is within line range
    pos = min(len(line), max(0, pos))

    if opt.get('lookAhead'):
        pos = offset_past_auto_closed(line, pos, opt)

    start = get_start_offset(line, pos, opt.get('prefix', ''))
    if start == -1:
        return None

    scanner = BackwardScanner(line, start)
    scanner.pos = pos
    stack = []

    while not scanner.sol():
        ch = scanner.peek()

        if Brackets.CurlyR in stack:
            if ch == Brackets.CurlyR:
                stack.append(ch)
                scanner.pos -= 1
                continue

            if ch != Brackets.CurlyL:
                scanner.pos -= 1
                continue

        if is_close_brace(ch, opt.get('type')):
            stack.append(ch)
        elif is_open_brace(ch, opt.get('type')):
            if not stack or stack.pop() != BRACE_PAIRS[ch]:
                # unexpected brace
                break
        elif Brackets.SquareR in stack or Brackets.CurlyR in stack:
            # respect all characters inside attribute sets or text nodes
            scanner.pos -= 1
            continue
        elif is_at_html_tag(scanner) or not is_abbreviation(ch):
            break

        scanner.pos -= 1

    if not stack and scanner.pos != pos:
        # Found something, remove some invalid symbols from the
        # beginning and return abbreviation
        abbreviation = re.sub(r'^[*+>^]+', '', line[scanner.pos:pos])
        prefix = opt.get('prefix', '')
        start = start - len(prefix) if prefix else pos - len(abbreviation)
        return ExtractedAbbreviation(abbreviation, pos - len(abbreviation), start, pos)


def offset_past_auto_closed(line: str, pos: int, options: dict):
    """
    Returns new `line` index which is right after characters beyond `pos` that
    editor will likely automatically close, e.g. }, ], and quotes
    """
    # closing quote is allowed only as a next character
    if pos < len(line) and is_quote(line[pos]): pos += 1

    # offset pointer until non-autoclosed character is found
    while pos < len(line) and is_close_brace(line[pos], options.get('type')):
        pos += 1

    return pos


def get_start_offset(line: str, pos: int, prefix: str):
    """
    Returns start offset (left limit) in `line` where we should stop looking for
    abbreviation: it’s nearest to `pos` location of `prefix` token
    """
    if not prefix: return 0

    scanner = BackwardScanner(line)
    scanner.pos = pos

    while not scanner.sol():
        if consume_pair(scanner, Brackets.SquareR, Brackets.SquareL) or consume_pair(scanner, Brackets.CurlyR, Brackets.CurlyL):
            continue

        result = scanner.pos
        if consume_list(scanner, prefix):
            return result

        scanner.pos -= 1

    return -1


def consume_pair(scanner: BackwardScanner, close_ch: str, open_ch: str):
    "Consumes full character pair, if possible"
    start = scanner.pos
    if scanner.consume(close_ch):
        while not scanner.sol():
            if scanner.consume(open_ch):
                return True

            scanner.pos -= 1

    scanner.pos = start
    return False


def consume_list(scanner: BackwardScanner, arr: list):
    "Consumes all character codes from given list, right-to-left, if possible"
    start = scanner.pos
    consumed = False
    i = len(arr) - 1

    while i >= 0 and not scanner.sol():
        if not scanner.consume(arr[i]):
            break
        consumed = i == 0
        i -= 1

    if not consumed:
        scanner.pos = start

    return consumed


def is_abbreviation(ch: str):
    return is_alpha(ch) or is_number(ch) or ch in SPECIAL_CHARS


def is_open_brace(ch: str, syntax: str):
    return ch == Brackets.RoundL or (syntax == 'markup' and ch in (Brackets.SquareL, Brackets.CurlyL))


def is_close_brace(ch: str, syntax: str):
    return ch == Brackets.RoundR or (syntax == 'markup' and ch in (Brackets.SquareR, Brackets.CurlyR))


def create_options(opt=None):
    options = {
        'type': 'markup',
        'lookAhead': True,
        'prefix': '',
    }

    if opt:
        options.update(opt)
    return options



class TemplatePlaceholder:
    __slots__ = ('before', 'after', 'name')

    def __init__(self, before: str, after: str, name: str):
        self.before = before
        self.after = after
        self.name = name


class TokenScanner:
    __slots__ = ('text', 'pos')

    def __init__(self, text: str):
        self.text = text
        self.pos = 0

    def peek(self, pos: int=None):
        if pos is None: pos = self.pos
        return self.text[pos] if pos < len(self.text) else ''


def template(text: str):
    """
    Splits given string into template tokens.
    Template is a string which contains placeholders which are uppercase names
    between `[` and `]`, for example: `[PLACEHOLDER]`.
    Unlike other templates, a placeholder may contain extra characters before and
    after name: `[%PLACEHOLDER.]`. If data for `PLACEHOLDER` is defined, it will
    be outputted with with these extra character, otherwise will be completely omitted.
    """
    tokens = []
    scanner = TokenScanner(text)
    offset = scanner.pos
    pos = scanner.pos

    while scanner.pos < len(scanner.text):
        pos = scanner.pos
        placeholder = consume_placeholder(scanner)
        if placeholder:
            if offset != scanner.pos:
                tokens.append(text[offset:pos])
            tokens.append(placeholder)
            offset = scanner.pos
        else:
            scanner.pos += 1

    if offset != scanner.pos:
        tokens.append(text[offset:])

    return tokens

def consume_placeholder(scanner: TokenScanner):
    if scanner.peek() == '[':
        scanner.pos += 1
        start = scanner.pos
        name_pos = start
        after_pos = start
        stack = 1

        while scanner.pos < len(scanner.text):
            ch = scanner.peek()
            if is_token_start(ch):
                name_pos = scanner.pos
                while is_token(scanner.peek()):
                    scanner.pos += 1
                after_pos = scanner.pos
            else:
                if ch == '[':
                    stack += 1
                elif ch == ']':
                    stack -= 1
                    if stack == 0:
                        pos = scanner.pos
                        scanner.pos += 1
                        return TemplatePlaceholder(
                            scanner.text[start:name_pos],
                            scanner.text[after_pos:pos],
                            scanner.text[name_pos:after_pos])

                scanner.pos += 1


def is_token_start(ch: str):
    return 'A' <= ch <= 'Z'


def is_token(ch: str):
    return is_token_start(ch) or ch == '_' or ch == '-' or '0' <= ch <= '9'



from ..scanner import Scanner
from ..scanner_utils import is_space, is_quote, is_alpha, is_number, eat_pair

class ElementType:
    Open = 1
    Close = 2
    SelfClose = 3

class Chars:
    Dash = '-'
    Dot = '.'
    Slash = '/'
    Colon = ':'
    LeftAngle = '<'
    RightAngle = '>'
    LeftRound = '('
    RightRound = ')'
    LeftSquare = '['
    RightSquare = ']'
    LeftCurly = '{'
    RightCurly = '}'
    Underscore = '_'
    Equals = '='
    Asterisk = '*'
    Hash = '#'

scan_opt = { 'throws': False }
default_special = {
    'style': None,
    'script': ['', 'text/javascript', 'application/x-javascript', 'javascript', 'typescript', 'ts', 'coffee', 'coffeescript']
}
default_empty = ['img', 'meta', 'link', 'br', 'base', 'hr', 'area', 'wbr', 'col', 'embed', 'input', 'param', 'source', 'track']

class ScannerOptions:
    __slots__ = ('xml', 'special', 'empty')

    def __init__(self, options: dict=None):
        if options is None:
            options = {}
        self.xml = options.get('xml', False)
        """
        Parses given source as XML document. It alters how should-be-empty
        elements are treated: for example, in XML mode parser will try to locate
        closing pair for `<br>` tag
        """

        self.special = options.get('special', default_special)
        """
        List of tags that should have special parsing rules, e.g. should not parse
        inner content and skip to closing tag. Key is a tag name that should be
        considered special and value is either empty (always mark element as special)
        or list of `type` attribute values, which, if present with one of this value,
        make element special
        """

        self.empty = options.get('empty', default_empty)
        """
        List of elements that should be treated as empty (e.g. without closing tag)
        in non-XML syntax
        """


def consume_array(scanner: Scanner, codes: list):
    "Consumes array of character codes from given scanner"
    start = scanner.pos

    for ch in codes:
        if not scanner.eat(ch):
            scanner.pos = start
            return False

    scanner.start = start
    return True


def consume_section(scanner: Scanner, prefix: list, suffix: list, allow_unclosed=False):
    """
    Consumes section from given string which starts with `open` character codes
    and ends with `close` character codes
    :return `true` if section was consumed
    """
    start = scanner.pos
    if consume_array(scanner, prefix):
        # consumed `<!--`, read next until we find ending part or reach the end of input
        while not scanner.eof():
            if consume_array(scanner, suffix):
                scanner.start = start
                return True

            scanner.pos += 1

        # unclosed section is allowed
        if allow_unclosed:
            scanner.start = start
            return True

        scanner.pos = start
        return False

    # unable to find section, revert to initial position
    scanner.pos = start
    return False


def name_start_char(ch: str):
    # Limited XML spec: https://www.w3.org/TR/xml/#NT-NameStartChar
    o = ord(ch) if ch else 0
    return is_alpha(ch) or \
        ch == Chars.Colon or \
        ch == Chars.Underscore or \
        0xC0  <= o <= 0xD6  or \
        0xD8  <= o <= 0xF6  or \
        0xF8  <= o <= 0x2FF or \
        0x370 <= o <= 0x37D or \
        0x37F <= o <= 0x1FFF


def name_char(ch: str):
    "Check if given character can be used in a tag or attribute name"
    # Limited XML spec: https://www.w3.org/TR/xml/#NT-NameChar
    o = ord(ch) if ch else 0
    return name_start_char(ch) or \
        ch == Chars.Dash or \
        ch == Chars.Dot or \
        is_number(ch) or \
        o == 0xB7 or \
        0x0300 <= o <= 0x036F


def ident(scanner: Scanner):
    "Consumes identifier from given scanner"
    start = scanner.pos
    if scanner.eat(name_start_char):
        scanner.eat_while(name_char)
        scanner.start = start
        return True

    return False


def is_terminator(ch: str):
    "Check if given code is tag terminator"
    return ch == Chars.RightAngle or ch == Chars.Slash


def is_unquoted(ch: str):
    "Check if given character code is valid unquoted value"
    return ch and not is_quote(ch) and not is_space(ch) and not is_terminator(ch)

def consume_paired(scanner: Scanner):
    """
    Consumes paired tokens (like `[` and `]`) with respect of nesting and embedded
    quoted values
    :return `true` if paired token was consumed
    """
    global scan_opt
    return eat_pair(scanner, Chars.LeftAngle, Chars.RightAngle, scan_opt) or \
        eat_pair(scanner, Chars.LeftRound, Chars.RightRound, scan_opt) or \
        eat_pair(scanner, Chars.LeftSquare, Chars.RightSquare, scan_opt) or \
        eat_pair(scanner, Chars.LeftCurly, Chars.RightCurly, scan_opt)


def get_unquoted_value(value: str):
    "Returns unquoted value of given string"
    # Trim quotes
    if value and is_quote(value[0]):
        value = value[1:]

    if is_quote(value[-1]):
        value = value[0:-1]

    return value



import unittest
import sys

sys.path.append('../')

from emmet.scanner import Scanner, ScannerException
from emmet.scanner_utils import eat_pair, eat_quoted

class TestScanner(unittest.TestCase):
    def test_basics(self):
        data = 'hello'
        s = Scanner(data)

        self.assertEqual(s.string, data)
        self.assertEqual(s.start, 0)
        self.assertEqual(s.pos, 0)

        self.assertEqual(s.peek(), data[0])
        self.assertEqual(s.start, 0)
        self.assertEqual(s.pos, 0)

        self.assertEqual(s.next(), data[0])
        self.assertEqual(s.next(), data[1])
        self.assertEqual(s.start, 0)
        self.assertEqual(s.pos, 2)

        self.assertEqual(s.next(), data[2])
        self.assertEqual(s.start, 0)
        self.assertEqual(s.pos, 3)

        self.assertEqual(s.current(), data[0:3])

    def test_limit_range(self):
        outer = Scanner('foo bar baz')
        inner = outer.limit(4, 7)

        self.assertTrue(outer != inner)

        outer_value = ''
        inner_value = ''

        while not outer.eof():
            outer_value += outer.next()

        while not inner.eof():
            inner_value += inner.next()

        self.assertEqual(outer_value, 'foo bar baz')
        self.assertEqual(inner_value, 'bar')

    def test_eat(self):
        scanner = Scanner('[foo] (bar (baz) bam)')

        self.assertTrue(eat_pair(scanner, '[', ']'))
        self.assertEqual(scanner.start, 0)
        self.assertEqual(scanner.pos, 5)
        self.assertEqual(scanner.current(), '[foo]')

        # No pair here
        self.assertFalse(eat_pair(scanner, '(', ')', { 'throws': True }))
        scanner.eat_while(' ')

        self.assertTrue(eat_pair(scanner, '(', ')', { 'throws': True }))
        self.assertEqual(scanner.start, 6)
        self.assertEqual(scanner.pos, 21)
        self.assertEqual(scanner.current(), '(bar (baz) bam)')

    def test_eat_with_quotes(self):
        scanner = Scanner('[foo "bar]" ]')
        self.assertTrue(eat_pair(scanner, '[', ']'))
        self.assertEqual(scanner.start, 0)
        self.assertEqual(scanner.pos, 13)
        self.assertEqual(scanner.current(), '[foo "bar]" ]')

    def test_eat_quoted(self):
        data = '"foo"   \'bar\''
        scanner = Scanner(data)

        self.assertTrue(eat_quoted(scanner))
        self.assertEqual(scanner.start, 0)
        self.assertEqual(scanner.pos, 5)
        self.assertEqual(scanner.current(), '"foo"')

        # no double-quoted value ahead
        self.assertFalse(eat_quoted(scanner, { 'throws': True }))

        # eat space
        self.assertTrue(scanner.eat_while(' '))
        self.assertEqual(scanner.pos, 8)

        self.assertTrue(eat_quoted(scanner))
        self.assertEqual(scanner.start, 8)
        self.assertEqual(scanner.pos, 13)
        self.assertEqual(scanner.current(), '\'bar\'')
        self.assertTrue(scanner.eof())

    def test_eat_invalid(self):
        scanner = Scanner('[foo')
        self.assertFalse(eat_pair(scanner, '[', ']'))
        self.assertEqual(scanner.start, 0)
        self.assertEqual(scanner.pos, 0)

        with self.assertRaises(ScannerException):
            eat_pair(scanner, '[', ']', { 'throws': True })

    def test_invalid_strings(self):
        scanner = Scanner('"foo')
        self.assertFalse(eat_quoted(scanner))
        self.assertEqual(scanner.pos, 0)

        with self.assertRaises(ScannerException):
            eat_quoted(scanner, { 'throws': True })

    def test_handle_escapes(self):
        scanner = Scanner('"foo\\"bar" baz')
        self.assertTrue(eat_quoted(scanner))
        self.assertEqual(scanner.start, 0)
        self.assertEqual(scanner.pos, 10)
        self.assertEqual(scanner.current(), '"foo\\"bar"')

if __name__ == '__main__':
    unittest.main()



from ..scanner import Scanner
from ..scanner_utils import is_space
from .scan import literal, Chars

# NB: no `Minus` operator, it must be handled differently
operators = '+/*,'

def split_value(value: str, offset=0):
    "Splits given CSS value into token list"
    start = -1
    expression = 0
    pos = 0
    result = []
    scanner = Scanner(value)

    while not scanner.eof():
        pos = scanner.pos
        if scanner.eat(is_space) or scanner.eat(is_operator) or is_minus_operator(scanner):
            # Use space as value delimiter but only if not in expression context,
            # e.g. `1 2` are distinct values but `(1 2)` not
            if not expression and start != -1:
                result.append((offset + start, offset + pos))
                start = -1
            scanner.eat_while(is_space)
        else:
            if start == -1:
                start = scanner.pos

            if scanner.eat(Chars.LeftRound):
                expression += 1
            elif scanner.eat(Chars.RightRound):
                expression -= 1
            elif not literal(scanner):
                scanner.pos += 1

    if start != -1 and start != scanner.pos:
        result.append((offset + start, offset + scanner.pos))

    return result


def is_operator(ch: str):
    "Check if given character is an operator"
    return ch in operators


def is_minus_operator(scanner: Scanner):
    "Check if current scanner state is at minus operator"
    # Minus operator is tricky since CSS supports dashes in keyword names like
    # `no-repeat`
    start = scanner.pos
    if scanner.eat('-') and scanner.eat(is_space):
        return True

    scanner.pos = start
    return False



from .walk import WalkState
from ...abbreviation import AbbreviationNode, AbbreviationAttribute
from ...abbreviation.tokenizer.tokens import Field
from ...config import Config
from ...output_stream import OutputStream, is_inline

caret = [Field('', 0)]
"Default caret token"

def is_snippet(node: AbbreviationNode):
    "Check if given node is a snippet: a node without name and attributes"
    return node and not node.name and not node.attributes if node else False

def is_inline_element(node: AbbreviationNode, config: Config):
    """
    Check if given node is inline-level element, e.g. element with explicitly
    defined node name
    """
    return is_inline(node, config) if node else False

def push_tokens(tokens: list, state: WalkState):
    out = state.out
    largest_index = -1

    for t in tokens:
        if isinstance(t, str):
            out.push_string(t)
        else:
            out.push_field(state.field + t.index, t.name)
            if t.index > largest_index:
                largest_index = t.index

    if largest_index != -1:
        state.field += largest_index + 1


def split_by_lines(tokens: list):
    """
    Splits given value token by lines: returns array where each entry is a token list
    for a single line
    """
    result = []
    line = []

    for t in tokens:
        if isinstance(t, str):
            lines = t.splitlines()
            line.append(lines.pop(0) if lines else '')
            while lines:
                result.append(line)
                line = [lines.pop(0) or '']
        else:
            line.append(t)

    if line:
        result.append(line)

    return result

def should_output_attribute(attr: AbbreviationAttribute):
    # In case if attribute is implied, check if it has a defined value:
    # either non-empty value or quoted empty value
    return not attr.implied or attr.value_type != 'raw' or attr.value



from ..scanner import Scanner
from ..scanner_utils import eat_quoted, is_space
from .utils import Chars, ident, consume_paired, scan_opt, is_unquoted, get_unquoted_value


class AttributeToken:
    __slots__ = ('name', 'value', 'name_start', 'name_end', 'value_start', 'value_end')

    def __init__(self, name: str, name_start: int, name_end: int, value: str=None, value_start: int=None, value_end: int=None):
        self.name = name
        self.name_start = name_start
        self.name_end = name_end
        self.value = value
        self.value_start = value_start
        self.value_end = value_end

    def to_json(self):
        json = {
            'name': self.name,
            'name_start': self.name_start,
            'name_end': self.name_end
        }

        if self.value is not None:
            json['value'] = self.value
            json['value_start'] = self.value_start
            json['value_end'] = self.value_end

        return json


def attributes(src: str, name: str=None):
    """
    Parses given string as list of HTML attributes.
    :param src A fragment to parse. If `name` argument is provided, it must be an
    opening tag (`<a foo="bar">`), otherwise it should be a fragment between element
    name and tag closing angle (`foo="bar"`)
    :param name Tag name
    """
    result = []
    start = 0
    end = len(src)
    if name:
        start = len(name) + 1
        end -= 2 if src.endswith('/>') else 1

    scanner = Scanner(src, start, end)

    while not scanner.eof():
        scanner.eat_while(is_space)
        if attribute_name(scanner):
            token = AttributeToken(scanner.current(), scanner.start, scanner.pos)

            if scanner.eat(Chars.Equals) and attribute_value(scanner):
                token.value = scanner.current()
                token.value_start = scanner.start
                token.value_end = scanner.pos

            result.append(token)
        else:
            # Do not break on invalid attributes: we are not validating parser
            scanner.pos += 1

    return result


def attribute_name(scanner: Scanner):
    "Consumes attribute name from given scanner context"
    start = scanner.pos
    if scanner.eat(Chars.Asterisk) or scanner.eat(Chars.Hash):
        # Angular-style directives: `<section *ngIf="showSection">`, `<video #movieplayer ...>`
        ident(scanner)
        scanner.start = start
        return True

    # Attribute name could be a regular name or expression:
    # React-style – `<div {...props}>`
    # Angular-style – `<div [ng-for]>` or `<div *ng-for>`
    return consume_paired(scanner) or ident(scanner)

def attribute_value(scanner: Scanner):
    "Consumes attribute value"
    # Supported attribute values are quoted, React-like expressions (`{foo}`)
    # or unquoted literals
    return eat_quoted(scanner, scan_opt) or consume_paired(scanner) or unquoted(scanner)


def get_attribute_value(attrs: list, name: str):
    "Returns clean (unquoted) value of `name` attribute"
    for attr in attrs:
        if attr.name == name:
            return attr.value and get_unquoted_value(attr.value)


def unquoted(scanner: Scanner):
    "Consumes unquoted value"
    start = scanner.pos
    if scanner.eat_while(is_unquoted):
        scanner.start = start
        return True

    return False



import unittest
import sys

sys.path.append('../')

from emmet.markup import parse, html, haml, pug, slim
from emmet.config import Config


def tabstops(index: int, placeholder: str, **kwargs):
    if placeholder:
        return '${%d:%s}' % (index, placeholder)
    return '${%d}' % index


default_config = Config()
field = Config({
    'options': {
        'output.field': tabstops
    }
})

def create_profile(options: dict):
    return Config({'options': options})


def output_html(abbr: str, config=default_config):
    return html(parse(abbr, config), config)


class TestHTMLFormat(unittest.TestCase):
    def test_basic(self):
        self.assertEqual(output_html('div>p'), '<div>\n\t<p></p>\n</div>')
        self.assertEqual(output_html('div>p*3'), '<div>\n\t<p></p>\n\t<p></p>\n\t<p></p>\n</div>')
        self.assertEqual(output_html('div#a>p.b*2>span'), '<div id="a">\n\t<p class="b"><span></span></p>\n\t<p class="b"><span></span></p>\n</div>')
        self.assertEqual(output_html('div>div>div'), '<div>\n\t<div>\n\t\t<div></div>\n\t</div>\n</div>')
        self.assertEqual(output_html('table>tr*2>td{item}*2'),
            '<table>\n\t<tr>\n\t\t<td>item</td>\n\t\t<td>item</td>\n\t</tr>\n\t<tr>\n\t\t<td>item</td>\n\t\t<td>item</td>\n\t</tr>\n</table>')

    def test_inline_elements(self):
        profile = create_profile({ 'output.inlineBreak': 3 })
        break_inline = create_profile({ 'output.inlineBreak': 1 })
        keep_inline = create_profile({ 'output.inlineBreak': 0 })
        xhtml = create_profile({ 'output.selfClosingStyle': 'xhtml' })

        self.assertEqual(output_html('div>a>b*3', xhtml), '<div>\n\t<a href="">\n\t\t<b></b>\n\t\t<b></b>\n\t\t<b></b>\n\t</a>\n</div>')

        self.assertEqual(output_html('p>i', profile), '<p><i></i></p>')
        self.assertEqual(output_html('p>i*2', profile), '<p><i></i><i></i></p>')
        self.assertEqual(output_html('p>i*2', break_inline), '<p>\n\t<i></i>\n\t<i></i>\n</p>')
        self.assertEqual(output_html('p>i*3', profile), '<p>\n\t<i></i>\n\t<i></i>\n\t<i></i>\n</p>')
        self.assertEqual(output_html('p>i*3', keep_inline), '<p><i></i><i></i><i></i></p>')

        self.assertEqual(output_html('i*2', profile), '<i></i><i></i>')
        self.assertEqual(output_html('i*3', profile), '<i></i>\n<i></i>\n<i></i>')
        self.assertEqual(output_html('i{a}+i{b}', profile), '<i>a</i><i>b</i>')

        self.assertEqual(output_html('img[src]/+p', xhtml), '<img src="" alt="" />\n<p></p>')
        self.assertEqual(output_html('div>img[src]/+p', xhtml), '<div>\n\t<img src="" alt="" />\n\t<p></p>\n</div>')
        self.assertEqual(output_html('div>p+img[src]/', xhtml), '<div>\n\t<p></p>\n\t<img src="" alt="" />\n</div>')
        self.assertEqual(output_html('div>p+img[src]/+p', xhtml), '<div>\n\t<p></p>\n\t<img src="" alt="" />\n\t<p></p>\n</div>')
        self.assertEqual(output_html('div>p+img[src]/*2+p', xhtml), '<div>\n\t<p></p>\n\t<img src="" alt="" /><img src="" alt="" />\n\t<p></p>\n</div>')
        self.assertEqual(output_html('div>p+img[src]/*3+p', xhtml), '<div>\n\t<p></p>\n\t<img src="" alt="" />\n\t<img src="" alt="" />\n\t<img src="" alt="" />\n\t<p></p>\n</div>')


    def test_generate_fields(self):
        self.assertEqual(output_html('a[href]', field), '<a href="${1}">${2}</a>')
        self.assertEqual(output_html('a[href]*2', field), '<a href="${1}">${2}</a><a href="${3}">${4}</a>')
        self.assertEqual(output_html('{${0} ${1:foo} ${2:bar}}*2', field), '${1} ${2:foo} ${3:bar}\n${4} ${5:foo} ${6:bar}')
        self.assertEqual(output_html('{${0} ${1:foo} ${2:bar}}*2'), ' foo bar\n foo bar')
        self.assertEqual(output_html('ul>li*2', field), '<ul>\n\t<li>${1}</li>\n\t<li>${2}</li>\n</ul>')
        self.assertEqual(output_html('div>img[src]/', field), '<div><img src="${1}" alt="${2}"></div>')


    def test_mixed_content(self):
        self.assertEqual(output_html('div{foo}'), '<div>foo</div>')
        self.assertEqual(output_html('div>{foo}'), '<div>foo</div>')
        self.assertEqual(output_html('div>{foo}+{bar}'), '<div>\n\tfoo\n\tbar\n</div>')
        self.assertEqual(output_html('div>{foo}+{bar}+p'), '<div>\n\tfoo\n\tbar\n\t<p></p>\n</div>')
        self.assertEqual(output_html('div>{foo}+{bar}+p+{foo}+{bar}+p'), '<div>\n\tfoo\n\tbar\n\t<p></p>\n\tfoo\n\tbar\n\t<p></p>\n</div>')
        self.assertEqual(output_html('div>{foo}+p+{bar}'), '<div>\n\tfoo\n\t<p></p>\n\tbar\n</div>')
        self.assertEqual(output_html('div>{foo}>p'), '<div>\n\tfoo\n\t<p></p>\n</div>')

        self.assertEqual(output_html('div>{<!-- ${0} -->}'), '<div><!--  --></div>')
        self.assertEqual(output_html('div>{<!-- ${0} -->}+p'), '<div>\n\t<!--  -->\n\t<p></p>\n</div>')
        self.assertEqual(output_html('div>p+{<!-- ${0} -->}'), '<div>\n\t<p></p>\n\t<!--  -->\n</div>')
        self.assertEqual(output_html('div>{<!-- ${0} -->}>p'), '<div>\n\t<!-- <p></p> -->\n</div>')
        self.assertEqual(output_html('div>{<!-- ${0} -->}*2>p'), '<div>\n\t<!-- <p></p> -->\n\t<!-- <p></p> -->\n</div>')

        self.assertEqual(output_html('div>{<!-- ${0} -->}>p*2'), '<div>\n\t<!-- \n\t<p></p>\n\t<p></p>\n\t-->\n</div>')
        self.assertEqual(output_html('div>{<!-- ${0} -->}*2>p*2'), '<div>\n\t<!-- \n\t<p></p>\n\t<p></p>\n\t-->\n\t<!-- \n\t<p></p>\n\t<p></p>\n\t-->\n</div>')

        self.assertEqual(output_html('div>{<!-- ${0} -->}>b'), '<div>\n\t<!-- <b></b> -->\n</div>')
        self.assertEqual(output_html('div>{<!-- ${0} -->}>b*2'), '<div>\n\t<!-- <b></b><b></b> -->\n</div>')
        self.assertEqual(output_html('div>{<!-- ${0} -->}>b*3'), '<div>\n\t<!-- \n\t<b></b>\n\t<b></b>\n\t<b></b>\n\t-->\n</div>')

        self.assertEqual(output_html('div>{<!-- ${0} -->}', field), '<div><!-- ${1} --></div>')
        self.assertEqual(output_html('div>{<!-- ${0} -->}>b', field), '<div>\n\t<!-- <b>${1}</b> -->\n</div>')

    def test_self_closing(self):
        xml_style = create_profile({ 'output.selfClosingStyle': 'xml' })
        html_style = create_profile({ 'output.selfClosingStyle': 'html' })
        xhtml_style = create_profile({ 'output.selfClosingStyle': 'xhtml' })

        self.assertEqual(output_html('img[src]/', html_style), '<img src="" alt="">')
        self.assertEqual(output_html('img[src]/', xhtml_style), '<img src="" alt="" />')
        self.assertEqual(output_html('img[src]/', xml_style), '<img src="" alt=""/>')
        self.assertEqual(output_html('div>img[src]/', xhtml_style), '<div><img src="" alt="" /></div>')


    def test_boolean_attributes(self):
        compact = create_profile({'output.compactBoolean': True})
        no_compact = create_profile({'output.compactBoolean': False})

        self.assertEqual(output_html('p[b.]', no_compact), '<p b="b"></p>')
        self.assertEqual(output_html('p[b.]', compact), '<p b></p>')
        self.assertEqual(output_html('p[contenteditable]', compact), '<p contenteditable></p>')
        self.assertEqual(output_html('p[contenteditable]', no_compact), '<p contenteditable="contenteditable"></p>')
        self.assertEqual(output_html('p[contenteditable=foo]', compact), '<p contenteditable="foo"></p>')


    def test_no_formatting(self):
        profile = create_profile({ 'output.format': False })
        self.assertEqual(output_html('div>p', profile), '<div><p></p></div>')
        self.assertEqual(output_html('div>{foo}+p+{bar}', profile), '<div>foo<p></p>bar</div>')
        self.assertEqual(output_html('div>{foo}>p', profile), '<div>foo<p></p></div>')
        self.assertEqual(output_html('div>{<!-- ${0} -->}>p', profile), '<div><!-- <p></p> --></div>')

    def test_format_specific_nodes(self):
        self.assertEqual(output_html('{<!DOCTYPE html>}+html>(head>meta[charset=${charset}]/+title{${1:Document}})+body', field),
                '<!DOCTYPE html>\n<html>\n<head>\n\t<meta charset="UTF-8">\n\t<title>${2:Document}</title>\n</head>\n<body>\n\t${3}\n</body>\n</html>')

    def test_comment(self):
        opt = Config({ 'options': { 'comment.enabled': True } })

        self.assertEqual(output_html('ul>li.item', opt), '<ul>\n\t<li class="item"></li>\n\t<!-- /.item -->\n</ul>')
        self.assertEqual(output_html('div>ul>li.item#foo', opt), '<div>\n\t<ul>\n\t\t<li class="item" id="foo"></li>\n\t\t<!-- /#foo.item -->\n\t</ul>\n</div>')

        opt.options['comment.after'] = ' { [%ID] }'
        self.assertEqual(output_html('div>ul>li.item#foo', opt), '<div>\n\t<ul>\n\t\t<li class="item" id="foo"></li> { %foo }\n\t</ul>\n</div>')


def output_haml(abbr: str, config=default_config):
    return haml(parse(abbr, config), config)

class TestHAMLFormat(unittest.TestCase):
    def test_basic(self):
        self.assertEqual(output_haml('div#header>ul.nav>li[title=test].nav-item*2'),
            '#header\n\t%ul.nav\n\t\t%li.nav-item(title="test") \n\t\t%li.nav-item(title="test") ')

        # https://github.com/emmetio/emmet/issues/446
        self.assertEqual(output_haml('li>a'), '%li\n\t%a(href="") ')

        self.assertEqual(output_haml('div#foo[data-n1=v1 title=test data-n2=v2].bar'),
            '#foo.bar(data-n1="v1" title="test" data-n2="v2") ')

        profile = create_profile({ 'output.compactBoolean': True })
        self.assertEqual(output_haml('input[disabled. foo title=test]/', profile), '%input(type="text" disabled foo="" title="test")/')

        profile = create_profile({ 'output.compactBoolean': False })
        self.assertEqual(output_haml('input[disabled. foo title=test]/', profile), '%input(type="text" disabled=true foo="" title="test")/')

    def test_nodes_with_text(self):
        self.assertEqual(output_haml('{Text 1}'), 'Text 1')
        self.assertEqual(output_haml('span{Text 1}'), '%span Text 1')
        self.assertEqual(output_haml('span{Text 1}>b{Text 2}'), '%span Text 1\n\t%b Text 2')
        self.assertEqual(output_haml('span{Text 1\nText 2}>b{Text 3}'), '%span\n\tText 1 |\n\tText 2 |\n\t%b Text 3')
        self.assertEqual(output_haml('div>span{Text 1\nText 2\nText 123}>b{Text 3}'), '%div\n\t%span\n\t\tText 1   |\n\t\tText 2   |\n\t\tText 123 |\n\t\t%b Text 3')

    def test_generate_fields(self):
        self.assertEqual(output_haml('a[href]', field), '%a(href="${1}") ${2}')
        self.assertEqual(output_haml('a[href]*2', field), '%a(href="${1}") ${2}\n%a(href="${3}") ${4}')
        self.assertEqual(output_haml('{${0} ${1:foo} ${2:bar}}*2', field), '${1} ${2:foo} ${3:bar}${4} ${5:foo} ${6:bar}')
        self.assertEqual(output_haml('{${0} ${1:foo} ${2:bar}}*2'), ' foo bar foo bar')
        self.assertEqual(output_haml('ul>li*2', field), '%ul\n\t%li ${1}\n\t%li ${2}')
        self.assertEqual(output_haml('div>img[src]/', field), '%div\n\t%img(src="${1}" alt="${2}")/')

def output_pug(abbr: str, config=default_config):
    return pug(parse(abbr, config), config)

class TestPUGFormat(unittest.TestCase):
    def test_basic(self):
        self.assertEqual(output_pug('div#header>ul.nav>li[title=test].nav-item*2'),
            '#header\n\tul.nav\n\t\tli.nav-item(title="test") \n\t\tli.nav-item(title="test") ')

        self.assertEqual(output_pug('div#foo[data-n1=v1 title=test data-n2=v2].bar'),
            '#foo.bar(data-n1="v1", title="test", data-n2="v2") ')

        self.assertEqual(output_pug('input[disabled. foo title=test]'), 'input(type="text", disabled, foo="", title="test")')
        # Use closing slash for XML output format
        self.assertEqual(output_pug('input[disabled. foo title=test]', create_profile({ 'output.selfClosingStyle': 'xml' })), 'input(type="text", disabled, foo="", title="test")/');


    def test_nodes_with_test(self):
        self.assertEqual(output_pug('{Text 1}'), 'Text 1')
        self.assertEqual(output_pug('span{Text 1}'), 'span Text 1')
        self.assertEqual(output_pug('span{Text 1}>b{Text 2}'), 'span Text 1\n\tb Text 2')
        self.assertEqual(output_pug('span{Text 1\nText 2}>b{Text 3}'), 'span\n\t| Text 1\n\t| Text 2\n\tb Text 3')
        self.assertEqual(output_pug('div>span{Text 1\nText 2}>b{Text 3}'), 'div\n\tspan\n\t\t| Text 1\n\t\t| Text 2\n\t\tb Text 3')

    def test_generate_fields(self):
        self.assertEqual(output_pug('a[href]', field), 'a(href="${1}") ${2}')
        self.assertEqual(output_pug('a[href]*2', field), 'a(href="${1}") ${2}\na(href="${3}") ${4}')
        self.assertEqual(output_pug('{${0} ${1:foo} ${2:bar}}*2', field), '${1} ${2:foo} ${3:bar}${4} ${5:foo} ${6:bar}')
        self.assertEqual(output_pug('{${0} ${1:foo} ${2:bar}}*2'), ' foo bar foo bar')
        self.assertEqual(output_pug('ul>li*2', field), 'ul\n\tli ${1}\n\tli ${2}')
        self.assertEqual(output_pug('div>img[src]/', field), 'div\n\timg(src="${1}", alt="${2}")')


def output_slim(abbr: str, config=default_config):
    return slim(parse(abbr, config), config)

class TestSlimFormat(unittest.TestCase):
    def test_basic(self):
        self.assertEqual(output_slim('div#header>ul.nav>li[title=test].nav-item*2'),
                '#header\n\tul.nav\n\t\tli.nav-item title="test" \n\t\tli.nav-item title="test" ')

        self.assertEqual(output_slim('div#foo[data-n1=v1 title=test data-n2=v2].bar'),
            '#foo.bar data-n1="v1" title="test" data-n2="v2" ')

    def test_nodes_with_text(self):
        self.assertEqual(output_slim('{Text 1}'), 'Text 1')
        self.assertEqual(output_slim('span{Text 1}'), 'span Text 1')
        self.assertEqual(output_slim('span{Text 1}>b{Text 2}'), 'span Text 1\n\tb Text 2')
        self.assertEqual(output_slim('span{Text 1\nText 2}>b{Text 3}'), 'span\n\t| Text 1\n\t| Text 2\n\tb Text 3')
        self.assertEqual(output_slim('div>span{Text 1\nText 2}>b{Text 3}'), 'div\n\tspan\n\t\t| Text 1\n\t\t| Text 2\n\t\tb Text 3')

    def test_generate_fields(self):
        self.assertEqual(output_slim('a[href]', field), 'a href="${1}" ${2}')
        self.assertEqual(output_slim('a[href]*2', field), 'a href="${1}" ${2}\na href="${3}" ${4}')
        self.assertEqual(output_slim('{${0} ${1:foo} ${2:bar}}*2', field), '${1} ${2:foo} ${3:bar}${4} ${5:foo} ${6:bar}')
        self.assertEqual(output_slim('{${0} ${1:foo} ${2:bar}}*2'), ' foo bar foo bar')
        self.assertEqual(output_slim('ul>li*2', field), 'ul\n\tli ${1}\n\tli ${2}')
        self.assertEqual(output_slim('div>img[src]/', field), 'div\n\timg src="${1}" alt="${2}"/')



from .reader import BackwardScanner
from .brackets import Brackets, BRACE_PAIRS
from ..scanner_utils import is_alpha, is_number, is_quote

class Chars:
    Tab = '\t'
    Space = ' '
    Dash = '-'
    Slash = '/'
    Colon = ':'
    Equals = '='
    AngleLeft = '<'
    AngleRight = '>'


def is_html(scanner: BackwardScanner):
    "Check if given reader’s current position points at the end of HTML tag"
    start = scanner.pos

    if not scanner.consume(Chars.AngleRight):
        return False

    ok = False
    scanner.consume(Chars.Slash) # possibly self-closed element

    while not scanner.sol():
        scanner.consume_while(is_white_space)

        if consume_ident(scanner):
            # ate identifier: could be a tag name, boolean attribute or unquoted
            # attribute value
            if scanner.consume(Chars.Slash):
                # either closing tag or invalid tag
                ok = scanner.consume(Chars.AngleLeft)
                break
            elif scanner.consume(Chars.AngleLeft):
                # opening tag
                ok = True
                break
            elif scanner.consume(is_white_space):
                # boolean attribute
                continue
            elif scanner.consume(Chars.Equals):
                # simple unquoted value or invalid attribute
                if consume_ident(scanner): continue
                break
            elif consume_attribute_with_unquoted_value(scanner):
                # identifier was a part of unquoted value
                ok = True
                break

            # invalid tag
            break

        if consume_attribute(scanner):
            continue

        break

    scanner.pos = start
    return ok


def consume_attribute(scanner: BackwardScanner):
    """
    Consumes HTML attribute from given string.
    Returns `True` if attribute was consumed.
    """
    return consume_attribute_with_quoted_value(scanner) or \
        consume_attribute_with_unquoted_value(scanner)


def consume_attribute_with_quoted_value(scanner: BackwardScanner):
    start = scanner.pos
    if consume_quoted(scanner) and scanner.consume(Chars.Equals) and consume_ident(scanner):
        return True

    scanner.pos = start
    return False


def consume_attribute_with_unquoted_value(scanner: BackwardScanner):
    start = scanner.pos
    stack = []

    while not scanner.sol():
        ch = scanner.peek()
        if is_close_bracket(ch):
            stack.append(ch)
        elif is_open_bracket(ch):
            if not stack or stack.pop() != BRACE_PAIRS[ch]:
                # Unexpected open bracket
                break
        elif not is_unquoted_value(ch):
            break
        scanner.pos -= 1


    if start != scanner.pos and scanner.consume(Chars.Equals) and consume_ident(scanner):
        return True

    scanner.pos = start
    return False


def consume_ident(scanner: BackwardScanner):
    "Consumes HTML identifier from stream"
    return scanner.consume_while(is_ident)


def consume_quoted(scanner: BackwardScanner):
    start = scanner.pos
    quote = scanner.previous()

    if is_quote(quote):
        while not scanner.sol():
            if scanner.previous() == quote and scanner.peek() != '\\':
                return True

    scanner.pos = start
    return False


def is_ident(ch: str):
    "Check if given character code belongs to HTML identifier"
    return ch == Chars.Colon or ch == Chars.Dash or is_alpha(ch) or is_number(ch)


def is_white_space(ch: str):
    "Check if given code is a whitespace"
    return ch == Chars.Space or ch == Chars.Tab


def is_unquoted_value(ch: str):
    "Check if given code may belong to unquoted attribute value"
    return ch and ch != Chars.Equals and not is_white_space(ch) and not is_quote(ch)


def is_open_bracket(ch: int):
    return ch in (Brackets.CurlyL, Brackets.RoundL, Brackets.SquareL)


def is_close_bracket(ch: int):
    return ch in (Brackets.CurlyR, Brackets.RoundR, Brackets.SquareR)



class Token:
    __slots__ = ('start', 'end')

    def __init__(self, start: int=None, end: int=None):
        self.start = start
        self.end = end

    @property
    def type(self):
        "Type of current token"
        return self.__class__.__name__

    def to_json(self):
        return dict([(k, self.__getattribute__(k)) for k in dir(self) if not k.startswith('__') and k != 'to_json'])

class Chars:
    Hash = '#'
    Dollar = '$'
    Dash = '-'
    Dot = '.'
    Colon = ':'
    Comma = ','
    Excl = '!'
    At = '@'
    Percent = '%'
    Underscore = '_'
    RoundBracketOpen = '('
    RoundBracketClose = ')'
    CurlyBracketOpen = '{'
    CurlyBracketClose = '}'
    Sibling = '+'
    SingleQuote = "'"
    DoubleQuote = '"'
    Transparent = 't'
    Slash = '/'

class OperatorType:
    Sibling = '+'
    Important = '!'
    ArgumentDelimiter = ','
    ValueDelimiter = '-'
    PropertyDelimiter = ':'

class Operator(Token):
    __slots__ = ('operator',)

    def __init__(self, operator: OperatorType, *args):
        super(Operator, self).__init__(*args)
        self.operator = operator


class Bracket(Token):
    __slots__ = ('open',)

    def __init__(self, is_open: bool, *args):
        super(Bracket, self).__init__(*args)
        self.open = is_open

class Literal(Token):
    __slots__ = ('value',)

    def __init__(self, value: str, *args):
        super(Literal, self).__init__(*args)
        self.value = value


class CustomProperty(Token):
    __slots__ = ('value', )

    def __init__(self, value: str, *args):
        super(CustomProperty, self).__init__(*args)
        self.value = value

class NumberValue(Token):
    __slots__ = ('value', 'raw_value', 'unit')

    def __init__(self, value: int, raw_value: str, unit='', *args):
        super(NumberValue, self).__init__(*args)
        self.value = value
        self.raw_value = raw_value
        self.unit = unit

class ColorValue(Token):
    __slots__ = ('r', 'g', 'b', 'a', 'raw')

    def __init__(self, r=0, g=0, b=0, a=None, raw='', *args):
        super(ColorValue, self).__init__(*args)
        self.r = r
        self.g = g
        self.b = b
        self.a = a if a is not None else 1
        self.raw = raw

class StringValue(Token):
    __slots__ = ('value', 'quote')

    def __init__(self, value: str, quote='', *args):
        super(StringValue, self).__init__(*args)
        self.value = value
        self.quote = quote

class Field(Token):
    __slots__ = ('name', 'index')

    def __init__(self, name: str, index: int=None, *args):
        super(Field, self).__init__(*args)
        self.index = index
        self.name = name

class WhiteSpace(Token): pass
